# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import six
import struct
from os_ken.lib import addrconv
from os_ken.lib import stringify
from os_ken.lib.packet import packet_base

# IEEE 802.1ag OpCode
CFM_CC_MESSAGE = 0x01
CFM_LOOPBACK_REPLY = 0x02
CFM_LOOPBACK_MESSAGE = 0x03
CFM_LINK_TRACE_REPLY = 0x04
CFM_LINK_TRACE_MESSAGE = 0x05

# IEEE 802.1ag TLV type
CFM_END_TLV = 0x00
CFM_SENDER_ID_TLV = 0x01
CFM_PORT_STATUS_TLV = 0x02
CFM_DATA_TLV = 0x03
CFM_INTERFACE_STATUS_TLV = 0x04
CFM_REPLY_INGRESS_TLV = 0x05
CFM_REPLY_EGRESS_TLV = 0x06
CFM_LTM_EGRESS_IDENTIFIER_TLV = 0x07
CFM_LTR_EGRESS_IDENTIFIER_TLV = 0x08
CFM_ORGANIZATION_SPECIFIC_TLV = 0x1f

# IEEE 802.1ag CFM version
CFM_VERSION = 0


class cfm(packet_base.PacketBase):
    """CFM (Connectivity Fault Management) Protocol header class.

    http://standards.ieee.org/getieee802/download/802.1ag-2007.pdf

    OpCode Field range assignments

    +---------------+--------------------------------------------------+
    | OpCode range  | CFM PDU or organization                          |
    +===============+==================================================+
    | 0             | Reserved for IEEE 802.1                          |
    +---------------+--------------------------------------------------+
    | 1             | Continuity Check Message (CCM)                   |
    +---------------+--------------------------------------------------+
    | 2             | Loopback Reply (LBR)                             |
    +---------------+--------------------------------------------------+
    | 3             | Loopback Message (LBM)                           |
    +---------------+--------------------------------------------------+
    | 4             | Linktrace Reply (LTR)                            |
    +---------------+--------------------------------------------------+
    | 5             | Linktrace Message (LTM)                          |
    +---------------+--------------------------------------------------+
    | 06 - 31       | Reserved for IEEE 802.1                          |
    +---------------+--------------------------------------------------+
    | 32 - 63       | Defined by ITU-T Y.1731                          |
    +---------------+--------------------------------------------------+
    | 64 - 255      | Reserved for IEEE 802.1.                         |
    +---------------+--------------------------------------------------+

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ============== ========================================
    Attribute      Description
    ============== ========================================
    op             CFM PDU
    ============== ========================================

    """
    _PACK_STR = '!B'
    _MIN_LEN = struct.calcsize(_PACK_STR)
    _CFM_OPCODE = {}
    _TYPE = {
        'ascii': [
            'ltm_orig_addr', 'ltm_targ_addr'
        ]
    }

    @staticmethod
    def register_cfm_opcode(type_):
        def _register_cfm_opcode(cls):
            cfm._CFM_OPCODE[type_] = cls
            return cls
        return _register_cfm_opcode

    def __init__(self, op=None):
        super(cfm, self).__init__()
        assert isinstance(op, operation)
        self.op = op

    @classmethod
    def parser(cls, buf):
        (opcode, ) = struct.unpack_from(cls._PACK_STR, buf, cls._MIN_LEN)
        cls_ = cls._CFM_OPCODE.get(opcode)
        op = cls_.parser(buf)
        instance = cls(op)
        rest = buf[len(instance):]
        return instance, None, rest

    def serialize(self, payload, prev):
        buf = self.op.serialize()
        return buf

    def __len__(self):
        return len(self.op)


@six.add_metaclass(abc.ABCMeta)
class operation(stringify.StringifyMixin):

    _TLV_TYPES = {}
    _END_TLV_LEN = 1

    @staticmethod
    def register_tlv_types(type_):
        def _register_tlv_types(cls):
            operation._TLV_TYPES[type_] = cls
            return cls
        return _register_tlv_types

    def __init__(self, md_lv, version, tlvs):
        self.md_lv = md_lv
        self.version = version
        tlvs = tlvs or []
        assert isinstance(tlvs, list)
        for tlv_ in tlvs:
            assert isinstance(tlv_, tlv)
        self.tlvs = tlvs

    @classmethod
    @abc.abstractmethod
    def parser(cls, buf):
        pass

    @abc.abstractmethod
    def serialize(self):
        pass

    @abc.abstractmethod
    def __len__(self):
        pass

    @classmethod
    def _parser_tlvs(cls, buf):
        offset = 0
        tlvs = []
        while True:
            (type_, ) = struct.unpack_from('!B', buf, offset)
            cls_ = cls._TLV_TYPES.get(type_)
            if not cls_:
                assert type_ is CFM_END_TLV
                break
            tlv_ = cls_.parser(buf[offset:])
            tlvs.append(tlv_)
            offset += len(tlv_)
        return tlvs

    @staticmethod
    def _serialize_tlvs(tlvs):
        buf = bytearray()
        for tlv_ in tlvs:
            buf.extend(tlv_.serialize())
        return buf

    def _calc_len(self, len_):
        for tlv_ in self.tlvs:
            len_ += len(tlv_)
        len_ += self._END_TLV_LEN
        return len_


@cfm.register_cfm_opcode(CFM_CC_MESSAGE)
class cc_message(operation):

    """CFM (IEEE Std 802.1ag-2007) Continuity Check Message (CCM)
    encoder/decoder class.

    This is used with os_ken.lib.packet.cfm.cfm.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ==================== =======================================
    Attribute            Description
    ==================== =======================================
    md_lv                Maintenance Domain Level.
    version              The protocol version number.
    rdi                  RDI bit.
    interval             CCM Interval.The default is 4 (1 frame/s)
    seq_num              Sequence Number.
    mep_id               Maintenance association End Point Identifier.
    md_name_format       Maintenance Domain Name Format.
                         The default is 4 (Character string)
    md_name_length       Maintenance Domain Name Length.
                         (0 means automatically-calculate
                         when encoding.)
    md_name              Maintenance Domain Name.
    short_ma_name_format Short MA Name Format.
                         The default is 2 (Character string)
    short_ma_name_length Short MA Name Format Length.
                         (0 means automatically-calculate
                         when encoding.)
    short_ma_name        Short MA Name.
    tlvs                 TLVs.
    ==================== =======================================
    """

    _PACK_STR = '!4BIHB'

    _MIN_LEN = struct.calcsize(_PACK_STR)
    _TLV_OFFSET = 70
    _MD_NAME_FORMAT_LEN = 1
    _MD_NAME_LENGTH_LEN = 1
    _SHORT_MA_NAME_FORMAT_LEN = 1
    _SHORT_MA_NAME_LENGTH_LEN = 1
    _MA_ID_LEN = 64

    # Maintenance Domain Name Format
    _MD_FMT_NO_MD_NAME_PRESENT = 1
    _MD_FMT_DOMAIN_NAME_BASED_STRING = 2
    _MD_FMT_MAC_ADDRESS_TWO_OCTET_INTEGER = 3
    _MD_FMT_CHARACTER_STRING = 4

    # Short MA Name Format
    _SHORT_MA_FMT_PRIMARY_VID = 1
    _SHORT_MA_FMT_CHARACTER_STRING = 2
    _SHORT_MA_FMT_TWO_OCTET_INTEGER = 3
    _SHORT_MA_FMT_RFC_2685_VPN_ID = 4

    # CCM Transmission Interval
    _INTERVAL_300_HZ = 1
    _INTERVAL_10_MSEC = 2
    _INTERVAL_100_MSEC = 3
    _INTERVAL_1_SEC = 4
    _INTERVAL_10_SEC = 5
    _INTERVAL_1_MIN = 6
    _INTERVAL_10_MIN = 6

    def __init__(self, md_lv=0, version=CFM_VERSION,
                 rdi=0, interval=_INTERVAL_1_SEC, seq_num=0, mep_id=1,
                 md_name_format=_MD_FMT_CHARACTER_STRING,
                 md_name_length=0, md_name=b"0",
                 short_ma_name_format=_SHORT_MA_FMT_CHARACTER_STRING,
                 short_ma_name_length=0, short_ma_name=b"1",
                 tlvs=None):
        super(cc_message, self).__init__(md_lv, version, tlvs)
        self._opcode = CFM_CC_MESSAGE
        assert rdi in [0, 1]
        self.rdi = rdi
        assert interval != 0
        self.interval = interval
        self.seq_num = seq_num
        assert 1 <= mep_id <= 8191
        self.mep_id = mep_id
        self.md_name_format = md_name_format
        self.md_name_length = md_name_length
        self.md_name = md_name
        self.short_ma_name_format = short_ma_name_format
        self.short_ma_name_length = short_ma_name_length
        self.short_ma_name = short_ma_name

    @classmethod
    def parser(cls, buf):
        (md_lv_version, opcode, flags, tlv_offset, seq_num, mep_id,
         md_name_format) = struct.unpack_from(cls._PACK_STR, buf)
        md_name_length = 0
        md_name = b""
        md_lv = int(md_lv_version >> 5)
        version = int(md_lv_version & 0x1f)
        rdi = int(flags >> 7)
        interval = int(flags & 0x07)
        offset = cls._MIN_LEN
        # parse md_name
        if md_name_format != cls._MD_FMT_NO_MD_NAME_PRESENT:
            (md_name_length, ) = struct.unpack_from("!B", buf, offset)
            offset += cls._MD_NAME_LENGTH_LEN
            form = "%dB" % md_name_length
            md_name = struct.unpack_from(form, buf, offset)
            offset += md_name_length
        # parse short_ma_name
        (short_ma_name_format,
         short_ma_name_length) = struct.unpack_from("!2B", buf, offset)
        offset += (cls._SHORT_MA_NAME_FORMAT_LEN +
                   cls._SHORT_MA_NAME_LENGTH_LEN)
        form = "%dB" % short_ma_name_length
        short_ma_name = struct.unpack_from(form, buf, offset)
        offset = cls._MIN_LEN + (cls._MA_ID_LEN - cls._MD_NAME_FORMAT_LEN)
        tlvs = cls._parser_tlvs(buf[offset:])
        # ascii to text
        if md_name_format == cls._MD_FMT_DOMAIN_NAME_BASED_STRING or \
           md_name_format == cls._MD_FMT_CHARACTER_STRING:
            md_name = b"".join(map(six.int2byte, md_name))
        if short_ma_name_format == cls._SHORT_MA_FMT_CHARACTER_STRING:
            short_ma_name = b"".join(map(six.int2byte, short_ma_name))
        return cls(md_lv, version, rdi, interval, seq_num, mep_id,
                   md_name_format, md_name_length,
                   md_name,
                   short_ma_name_format, short_ma_name_length,
                   short_ma_name,
                   tlvs)

    def serialize(self):
        buf = struct.pack(self._PACK_STR,
                          (self.md_lv << 5) | self.version,
                          self._opcode,
                          (self.rdi << 7) | self.interval,
                          self._TLV_OFFSET,
                          self.seq_num, self.mep_id, self.md_name_format)
        buf = bytearray(buf)
        # Maintenance Domain Name
        if self.md_name_format != self._MD_FMT_NO_MD_NAME_PRESENT:
            if self.md_name_length == 0:
                self.md_name_length = len(self.md_name)
            buf.extend(struct.pack('!B%ds' % self.md_name_length,
                                   self.md_name_length, self.md_name))
        # Short MA Name
        if self.short_ma_name_length == 0:
            self.short_ma_name_length = len(self.short_ma_name)
        buf.extend(struct.pack('!2B%ds' % self.short_ma_name_length,
                               self.short_ma_name_format,
                               self.short_ma_name_length,
                               self.short_ma_name
                               ))
        # 0 pad
        maid_length = (self._MD_NAME_FORMAT_LEN +
                       self._SHORT_MA_NAME_FORMAT_LEN +
                       self._SHORT_MA_NAME_LENGTH_LEN +
                       self.short_ma_name_length)
        if self.md_name_format != self._MD_FMT_NO_MD_NAME_PRESENT:
            maid_length += self._MD_NAME_LENGTH_LEN + self.md_name_length
        buf.extend(bytearray(self._MA_ID_LEN - maid_length))
        # tlvs
        if self.tlvs:
            buf.extend(self._serialize_tlvs(self.tlvs))
        buf.extend(struct.pack("!B", CFM_END_TLV))
        return buf

    def __len__(self):
        return self._calc_len(
            (self._MIN_LEN - self._MD_NAME_FORMAT_LEN) + self._MA_ID_LEN)


class loopback(operation):

    _PACK_STR = '!4BI'
    _MIN_LEN = struct.calcsize(_PACK_STR)
    _TLV_OFFSET = 4

    @abc.abstractmethod
    def __init__(self, md_lv, version, transaction_id, tlvs):
        super(loopback, self).__init__(md_lv, version, tlvs)
        self._flags = 0
        self.transaction_id = transaction_id

    @classmethod
    def parser(cls, buf):
        (md_lv_version, opcode, flags, tlv_offset,
         transaction_id) = struct.unpack_from(cls._PACK_STR, buf)
        md_lv = int(md_lv_version >> 5)
        version = int(md_lv_version & 0x1f)
        tlvs = cls._parser_tlvs(buf[cls._MIN_LEN:])
        return cls(md_lv, version, transaction_id, tlvs)

    def serialize(self):
        buf = struct.pack(self._PACK_STR,
                          (self.md_lv << 5) | self.version,
                          self._opcode,
                          self._flags,
                          self._TLV_OFFSET,
                          self.transaction_id,
                          )
        buf = bytearray(buf)
        # tlvs
        if self.tlvs:
            buf.extend(self._serialize_tlvs(self.tlvs))
        buf.extend(struct.pack("!B", CFM_END_TLV))

        return buf

    def __len__(self):
        return self._calc_len(self._MIN_LEN)


@cfm.register_cfm_opcode(CFM_LOOPBACK_MESSAGE)
class loopback_message(loopback):

    """CFM (IEEE Std 802.1ag-2007) Loopback Message (LBM) encoder/decoder class.

    This is used with os_ken.lib.packet.cfm.cfm.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ================= =======================================
    Attribute         Description
    ================= =======================================
    md_lv             Maintenance Domain Level.
    version           The protocol version number.
    transaction_id    Loopback Transaction Identifier.
    tlvs              TLVs.
    ================= =======================================
    """

    def __init__(self, md_lv=0, version=CFM_VERSION,
                 transaction_id=0,
                 tlvs=None,
                 ):
        super(loopback_message, self).__init__(md_lv, version,
                                               transaction_id,
                                               tlvs)
        self._opcode = CFM_LOOPBACK_MESSAGE


@cfm.register_cfm_opcode(CFM_LOOPBACK_REPLY)
class loopback_reply(loopback):

    """CFM (IEEE Std 802.1ag-2007) Loopback Reply (LBR) encoder/decoder class.

    This is used with os_ken.lib.packet.cfm.cfm.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ==================== =======================================
    Attribute            Description
    ==================== =======================================
    md_lv                Maintenance Domain Level.
    version              The protocol version number.
    transaction_id       Loopback Transaction Identifier.
    tlvs                 TLVs.
    ==================== =======================================
    """

    def __init__(self, md_lv=0, version=CFM_VERSION,
                 transaction_id=0,
                 tlvs=None,
                 ):
        super(loopback_reply, self).__init__(md_lv, version,
                                             transaction_id,
                                             tlvs)
        self._opcode = CFM_LOOPBACK_REPLY


class link_trace(operation):

    @abc.abstractmethod
    def __init__(self, md_lv, version, use_fdb_only,
                 transaction_id, ttl, tlvs):
        super(link_trace, self).__init__(md_lv, version, tlvs)
        assert use_fdb_only in [0, 1]
        self.use_fdb_only = use_fdb_only
        self.transaction_id = transaction_id
        self.ttl = ttl

    @classmethod
    @abc.abstractmethod
    def parser(cls, buf):
        pass

    @abc.abstractmethod
    def serialize(self):
        pass

    def __len__(self):
        return self._calc_len(self._MIN_LEN)


@cfm.register_cfm_opcode(CFM_LINK_TRACE_MESSAGE)
class link_trace_message(link_trace):

    """CFM (IEEE Std 802.1ag-2007) Linktrace Message (LTM)
    encoder/decoder class.

    This is used with os_ken.lib.packet.cfm.cfm.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ==================== =======================================
    Attribute            Description
    ==================== =======================================
    md_lv                Maintenance Domain Level.
    version              The protocol version number.
    use_fdb_only         UseFDBonly bit.
    transaction_id       LTM Transaction Identifier.
    ttl                  LTM TTL.
    ltm_orig_addr        Original MAC Address.
    ltm_targ_addr        Target MAC Address.
    tlvs                 TLVs.
    ==================== =======================================
    """

    _PACK_STR = '!4BIB6s6s'
    _ALL_PACK_LEN = struct.calcsize(_PACK_STR)
    _MIN_LEN = _ALL_PACK_LEN
    _TLV_OFFSET = 17
    _TYPE = {
        'ascii': [
            'ltm_orig_addr', 'ltm_targ_addr'
        ]
    }

    def __init__(self, md_lv=0, version=CFM_VERSION,
                 use_fdb_only=1,
                 transaction_id=0,
                 ttl=64,
                 ltm_orig_addr='00:00:00:00:00:00',
                 ltm_targ_addr='00:00:00:00:00:00',
                 tlvs=None
                 ):
        super(link_trace_message, self).__init__(md_lv, version,
                                                 use_fdb_only,
                                                 transaction_id,
                                                 ttl,
                                                 tlvs)
        self._opcode = CFM_LINK_TRACE_MESSAGE
        self.ltm_orig_addr = ltm_orig_addr
        self.ltm_targ_addr = ltm_targ_addr

    @classmethod
    def parser(cls, buf):
        (md_lv_version, opcode, flags, tlv_offset, transaction_id, ttl,
         ltm_orig_addr, ltm_targ_addr) = struct.unpack_from(cls._PACK_STR, buf)
        md_lv = int(md_lv_version >> 5)
        version = int(md_lv_version & 0x1f)
        use_fdb_only = int(flags >> 7)
        tlvs = cls._parser_tlvs(buf[cls._MIN_LEN:])
        return cls(md_lv, version, use_fdb_only,
                   transaction_id, ttl,
                   addrconv.mac.bin_to_text(ltm_orig_addr),
                   addrconv.mac.bin_to_text(ltm_targ_addr),
                   tlvs)

    def serialize(self):
        buf = struct.pack(self._PACK_STR,
                          (self.md_lv << 5) | self.version,
                          self._opcode,
                          self.use_fdb_only << 7,
                          self._TLV_OFFSET,
                          self.transaction_id,
                          self.ttl,
                          addrconv.mac.text_to_bin(self.ltm_orig_addr),
                          addrconv.mac.text_to_bin(self.ltm_targ_addr),
                          )
        buf = bytearray(buf)
        if self.tlvs:
            buf.extend(self._serialize_tlvs(self.tlvs))
        buf.extend(struct.pack("!B", CFM_END_TLV))
        return buf


@cfm.register_cfm_opcode(CFM_LINK_TRACE_REPLY)
class link_trace_reply(link_trace):

    """CFM (IEEE Std 802.1ag-2007) Linktrace Reply (LTR) encoder/decoder class.

    This is used with os_ken.lib.packet.cfm.cfm.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ==================== =======================================
    Attribute            Description
    ==================== =======================================
    version              The protocol version number.
    use_fdb_only         UseFDBonly bit.
    fwd_yes              FwdYes bit.
    terminal_mep         TerminalMep bit.
    transaction_id       LTR Transaction Identifier.
    ttl                  Reply TTL.
    relay_action         Relay Action.The default is 1 (RlyHit)
    tlvs                 TLVs.
    ==================== =======================================
    """
    _PACK_STR = '!4BIBB'
    _ALL_PACK_LEN = struct.calcsize(_PACK_STR)
    _MIN_LEN = _ALL_PACK_LEN
    _TLV_OFFSET = 6

    # Relay Action field values
    _RLY_HIT = 1
    _RLY_FDB = 2
    _RLY_MPDB = 3

    def __init__(self, md_lv=0, version=CFM_VERSION, use_fdb_only=1,
                 fwd_yes=0, terminal_mep=1, transaction_id=0,
                 ttl=64, relay_action=_RLY_HIT, tlvs=None
                 ):
        super(link_trace_reply, self).__init__(md_lv, version,
                                               use_fdb_only,
                                               transaction_id,
                                               ttl,
                                               tlvs)
        self._opcode = CFM_LINK_TRACE_REPLY
        assert fwd_yes in [0, 1]
        self.fwd_yes = fwd_yes
        assert terminal_mep in [0, 1]
        self.terminal_mep = terminal_mep
        assert relay_action in [self._RLY_HIT, self._RLY_FDB, self._RLY_MPDB]
        self.relay_action = relay_action

    @classmethod
    def parser(cls, buf):
        (md_lv_version, opcode, flags, tlv_offset, transaction_id, ttl,
         relay_action) = struct.unpack_from(cls._PACK_STR, buf)
        md_lv = int(md_lv_version >> 5)
        version = int(md_lv_version & 0x1f)
        use_fdb_only = int(flags >> 7)
        fwd_yes = int(flags >> 6 & 0x01)
        terminal_mep = int(flags >> 5 & 0x01)
        tlvs = cls._parser_tlvs(buf[cls._MIN_LEN:])
        return cls(md_lv, version, use_fdb_only, fwd_yes, terminal_mep,
                   transaction_id, ttl, relay_action, tlvs)

    def serialize(self):
        buf = struct.pack(self._PACK_STR,
                          (self.md_lv << 5) | self.version,
                          self._opcode,
                          (self.use_fdb_only << 7) |
                          (self.fwd_yes << 6) |
                          (self.terminal_mep << 5),
                          self._TLV_OFFSET,
                          self.transaction_id,
                          self.ttl,
                          self.relay_action,
                          )
        buf = bytearray(buf)
        if self.tlvs:
            buf.extend(self._serialize_tlvs(self.tlvs))
        buf.extend(struct.pack("!B", CFM_END_TLV))
        return buf


cfm.set_classes(cfm._CFM_OPCODE)


@six.add_metaclass(abc.ABCMeta)
class tlv(stringify.StringifyMixin):

    _TYPE_LEN = 1
    _LENGTH_LEN = 2
    _TYPE = {
        'ascii': [
            'egress_id_mac', 'last_egress_id_mac',
            'next_egress_id_mac', 'mac_address'
        ]
    }

    def __init__(self, length):
        self.length = length

    @classmethod
    @abc.abstractmethod
    def parser(cls, buf):
        pass

    @abc.abstractmethod
    def serialize(self):
        pass

    def __len__(self):
        return self.length + self._TYPE_LEN + self._LENGTH_LEN


@operation.register_tlv_types(CFM_SENDER_ID_TLV)
class sender_id_tlv(tlv):

    """CFM (IEEE Std 802.1ag-2007) Sender ID TLV encoder/decoder class.

    This is used with os_ken.lib.packet.cfm.cfm.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ==================== =======================================
    Attribute            Description
    ==================== =======================================
    length               Length of Value field.
                         (0 means automatically-calculate when encoding.)
    chassis_id_length    Chassis ID Length.
                         (0 means automatically-calculate when encoding.)
    chassis_id_subtype   Chassis ID Subtype.
                         The default is 4 (Mac Address)
    chassis_id           Chassis ID.
    ma_domain_length     Management Address Domain Length.
                         (0 means automatically-calculate when encoding.)
    ma_domain            Management Address Domain.
    ma_length            Management Address Length.
                         (0 means automatically-calculate when encoding.)
    ma                   Management Address.
    ==================== =======================================
    """

    _PACK_STR = '!BHB'
    _MIN_LEN = struct.calcsize(_PACK_STR)
    _CHASSIS_ID_LENGTH_LEN = 1
    _CHASSIS_ID_SUBTYPE_LEN = 1
    _MA_DOMAIN_LENGTH_LEN = 1
    _MA_LENGTH_LEN = 1

    # Chassis ID subtype enumeration
    _CHASSIS_ID_CHASSIS_COMPONENT = 1
    _CHASSIS_ID_INTERFACE_ALIAS = 2
    _CHASSIS_ID_PORT_COMPONENT = 3
    _CHASSIS_ID_MAC_ADDRESS = 4
    _CHASSIS_ID_NETWORK_ADDRESS = 5
    _CHASSIS_ID_INTERFACE_NAME = 6
    _CHASSIS_ID_LOCALLY_ASSIGNED = 7

    def __init__(self,
                 length=0,
                 chassis_id_length=0,
                 chassis_id_subtype=_CHASSIS_ID_MAC_ADDRESS,
                 chassis_id=b'',
                 ma_domain_length=0,
                 ma_domain=b'',
                 ma_length=0,
                 ma=b''
                 ):
        super(sender_id_tlv, self).__init__(length)
        self._type = CFM_SENDER_ID_TLV
        self.chassis_id_length = chassis_id_length
        assert chassis_id_subtype in [
            self._CHASSIS_ID_CHASSIS_COMPONENT,
            self._CHASSIS_ID_INTERFACE_ALIAS,
            self._CHASSIS_ID_PORT_COMPONENT,
            self._CHASSIS_ID_MAC_ADDRESS,
            self._CHASSIS_ID_NETWORK_ADDRESS,
            self._CHASSIS_ID_INTERFACE_NAME,
            self._CHASSIS_ID_LOCALLY_ASSIGNED]
        self.chassis_id_subtype = chassis_id_subtype
        self.chassis_id = chassis_id
        self.ma_domain_length = ma_domain_length
        self.ma_domain = ma_domain
        self.ma_length = ma_length
        self.ma = ma

    @classmethod
    def parser(cls, buf):
        (type_, length, chassis_id_length) = struct.unpack_from(cls._PACK_STR,
                                                                buf)
        chassis_id_subtype = 4
        chassis_id = b''
        ma_domain_length = 0
        ma_domain = b''
        ma_length = 0
        ma = b''
        offset = cls._MIN_LEN
        if chassis_id_length != 0:
            (chassis_id_subtype, ) = struct.unpack_from("!B", buf, offset)
            offset += cls._CHASSIS_ID_SUBTYPE_LEN
            form = "%ds" % chassis_id_length
            (chassis_id,) = struct.unpack_from(form, buf, offset)
            offset += chassis_id_length
        if length + (cls._TYPE_LEN + cls._LENGTH_LEN) > offset:
            (ma_domain_length, ) = struct.unpack_from("!B", buf, offset)
            offset += cls._MA_DOMAIN_LENGTH_LEN
            form = "%ds" % ma_domain_length
            (ma_domain, ) = struct.unpack_from(form, buf, offset)
            offset += ma_domain_length
            if length + (cls._TYPE_LEN + cls._LENGTH_LEN) > offset:
                (ma_length, ) = struct.unpack_from("!B", buf, offset)
                offset += cls._MA_LENGTH_LEN
                form = "%ds" % ma_length
                (ma, ) = struct.unpack_from(form, buf, offset)
        return cls(length, chassis_id_length, chassis_id_subtype,
                   chassis_id, ma_domain_length, ma_domain, ma_length, ma)

    def serialize(self):
        # calculate length when it contains 0
        if self.chassis_id_length == 0:
            self.chassis_id_length = len(self.chassis_id)
        if self.ma_domain_length == 0:
            self.ma_domain_length = len(self.ma_domain)
        if self.ma_length == 0:
            self.ma_length = len(self.ma)
        if self.length == 0:
            self.length += self._CHASSIS_ID_LENGTH_LEN
            if self.chassis_id_length != 0:
                self.length += (self._CHASSIS_ID_SUBTYPE_LEN +
                                self.chassis_id_length)
            if self.chassis_id_length != 0 or self.ma_domain_length != 0:
                self.length += self._MA_DOMAIN_LENGTH_LEN
            if self.ma_domain_length != 0:
                self.length += (self.ma_domain_length +
                                self._MA_LENGTH_LEN + self.ma_length)
        # start serialize
        buf = struct.pack(self._PACK_STR,
                          self._type,
                          self.length,
                          self.chassis_id_length
                          )
        buf = bytearray(buf)
        # Chassis ID Subtype and Chassis ID present
        # if the Chassis ID Length field contains not 0.
        if self.chassis_id_length != 0:
            buf.extend(struct.pack("!B", self.chassis_id_subtype))
            form = "%ds" % self.chassis_id_length
            buf.extend(struct.pack(form, self.chassis_id))
        # Management Address Domain Length present
        # if the Chassis ID Length field or Management Address Length field
        # contains not 0.
        if self.chassis_id_length != 0 or self.ma_domain_length != 0:
            buf.extend(struct.pack("!B", self.ma_domain_length))
        # Management Address Domain present
        # Management Address Domain Length field contains not 0.
        if self.ma_domain_length != 0:
            form = "%ds" % self.ma_domain_length
            buf.extend(struct.pack(form, self.ma_domain))
            buf.extend(struct.pack("!B", self.ma_length))
            # Management Address present
            # Management Address Length field contains not 0.
            if self.ma_length != 0:
                form = "%ds" % self.ma_length
                buf.extend(struct.pack(form, self.ma))
        return buf


@operation.register_tlv_types(CFM_PORT_STATUS_TLV)
class port_status_tlv(tlv):

    """CFM (IEEE Std 802.1ag-2007) Port Status TLV encoder/decoder class.

    This is used with os_ken.lib.packet.cfm.cfm.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ==================== =======================================
    Attribute            Description
    ==================== =======================================
    length               Length of Value field.
                         (0 means automatically-calculate when encoding.)
    port_status          Port Status.The default is 1 (psUp)
    ==================== =======================================
    """

    _PACK_STR = '!BHB'
    _MIN_LEN = struct.calcsize(_PACK_STR)

    # Port Status TLV values
    _PS_BLOCKED = 1
    _PS_UP = 2

    def __init__(self, length=0, port_status=_PS_UP):
        super(port_status_tlv, self).__init__(length)
        self._type = CFM_PORT_STATUS_TLV
        assert port_status in [self._PS_BLOCKED, self._PS_UP]
        self.port_status = port_status

    @classmethod
    def parser(cls, buf):
        (type_, length,
         port_status) = struct.unpack_from(cls._PACK_STR, buf)
        return cls(length, port_status)

    def serialize(self):
        # calculate length when it contains 0
        if self.length == 0:
            self.length = 1
        # start serialize
        buf = struct.pack(self._PACK_STR,
                          self._type,
                          self.length,
                          self.port_status)
        return bytearray(buf)


@operation.register_tlv_types(CFM_DATA_TLV)
class data_tlv(tlv):

    """CFM (IEEE Std 802.1ag-2007) Data TLV encoder/decoder class.

    This is used with os_ken.lib.packet.cfm.cfm.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ============== =======================================
    Attribute      Description
    ============== =======================================
    length         Length of Value field.
                   (0 means automatically-calculate when encoding)
    data_value     Bit pattern of any of n octets.(n = length)
    ============== =======================================
    """

    _PACK_STR = '!BH'
    _MIN_LEN = struct.calcsize(_PACK_STR)

    def __init__(self, length=0, data_value=b""
                 ):
        super(data_tlv, self).__init__(length)
        self._type = CFM_DATA_TLV
        self.data_value = data_value

    @classmethod
    def parser(cls, buf):
        (type_, length) = struct.unpack_from(cls._PACK_STR, buf)
        form = "%ds" % length
        (data_value, ) = struct.unpack_from(form, buf, cls._MIN_LEN)
        return cls(length, data_value)

    def serialize(self):
        # calculate length when it contains 0
        if self.length == 0:
            self.length = len(self.data_value)
        # start serialize
        buf = struct.pack(self._PACK_STR,
                          self._type,
                          self.length)
        buf = bytearray(buf)
        form = "%ds" % self.length
        buf.extend(struct.pack(form, self.data_value))
        return buf


@operation.register_tlv_types(CFM_INTERFACE_STATUS_TLV)
class interface_status_tlv(tlv):

    """CFM (IEEE Std 802.1ag-2007) Interface Status TLV encoder/decoder class.

    This is used with os_ken.lib.packet.cfm.cfm.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ==================== =======================================
    Attribute            Description
    ==================== =======================================
    length               Length of Value field.
                         (0 means automatically-calculate when encoding.)
    interface_status     Interface Status.The default is 1 (isUp)
    ==================== =======================================
    """

    _PACK_STR = '!BHB'
    _MIN_LEN = struct.calcsize(_PACK_STR)

    # Interface Status TLV values
    _IS_UP = 1
    _IS_DOWN = 2
    _IS_TESTING = 3
    _IS_UNKNOWN = 4
    _IS_DORMANT = 5
    _IS_NOT_PRESENT = 6
    _IS_LOWER_LAYER_DOWN = 7

    def __init__(self, length=0, interface_status=_IS_UP):
        super(interface_status_tlv, self).__init__(length)
        self._type = CFM_INTERFACE_STATUS_TLV
        assert interface_status in [
            self._IS_UP, self._IS_DOWN, self._IS_TESTING,
            self._IS_UNKNOWN, self._IS_DORMANT, self._IS_NOT_PRESENT,
            self._IS_LOWER_LAYER_DOWN]
        self.interface_status = interface_status

    @classmethod
    def parser(cls, buf):
        (type_, length,
         interface_status) = struct.unpack_from(cls._PACK_STR, buf)
        return cls(length, interface_status)

    def serialize(self):
        # calculate length when it contains 0
        if self.length == 0:
            self.length = 1
        # start serialize
        buf = struct.pack(self._PACK_STR,
                          self._type,
                          self.length,
                          self.interface_status)
        return bytearray(buf)


@operation.register_tlv_types(CFM_LTM_EGRESS_IDENTIFIER_TLV)
class ltm_egress_identifier_tlv(tlv):

    """CFM (IEEE Std 802.1ag-2007) LTM EGRESS TLV encoder/decoder class.

    This is used with os_ken.lib.packet.cfm.cfm.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ============== =======================================
    Attribute      Description
    ============== =======================================
    length         Length of Value field.
                   (0 means automatically-calculate when encoding.)
    egress_id_ui   Egress Identifier of Unique ID.
    egress_id_mac  Egress Identifier of MAC address.
    ============== =======================================
    """

    _PACK_STR = '!BHH6s'
    _MIN_LEN = struct.calcsize(_PACK_STR)

    def __init__(self,
                 length=0,
                 egress_id_ui=0,
                 egress_id_mac='00:00:00:00:00:00'
                 ):
        super(ltm_egress_identifier_tlv, self).__init__(length)
        self._type = CFM_LTM_EGRESS_IDENTIFIER_TLV
        self.egress_id_ui = egress_id_ui
        self.egress_id_mac = egress_id_mac

    @classmethod
    def parser(cls, buf):
        (type_, length, egress_id_ui,
         egress_id_mac) = struct.unpack_from(cls._PACK_STR, buf)
        return cls(length, egress_id_ui,
                   addrconv.mac.bin_to_text(egress_id_mac))

    def serialize(self):
        # calculate length when it contains 0
        if self.length == 0:
            self.length = 8
        # start serialize
        buf = struct.pack(self._PACK_STR,
                          self._type,
                          self.length,
                          self.egress_id_ui,
                          addrconv.mac.text_to_bin(self.egress_id_mac)
                          )
        return bytearray(buf)


@operation.register_tlv_types(CFM_LTR_EGRESS_IDENTIFIER_TLV)
class ltr_egress_identifier_tlv(tlv):

    """CFM (IEEE Std 802.1ag-2007) LTR EGRESS TLV encoder/decoder class.

    This is used with os_ken.lib.packet.cfm.cfm.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ==================== =======================================
    Attribute            Description
    ==================== =======================================
    length               Length of Value field.
                         (0 means automatically-calculate when encoding.)
    last_egress_id_ui    Last Egress Identifier of Unique ID.
    last_egress_id_mac   Last Egress Identifier of MAC address.
    next_egress_id_ui    Next Egress Identifier of Unique ID.
    next_egress_id_mac   Next Egress Identifier of MAC address.
    ==================== =======================================
    """

    _PACK_STR = '!BHH6sH6s'
    _MIN_LEN = struct.calcsize(_PACK_STR)

    def __init__(self,
                 length=0,
                 last_egress_id_ui=0,
                 last_egress_id_mac='00:00:00:00:00:00',
                 next_egress_id_ui=0,
                 next_egress_id_mac='00:00:00:00:00:00'
                 ):
        super(ltr_egress_identifier_tlv, self).__init__(length)
        self._type = CFM_LTR_EGRESS_IDENTIFIER_TLV
        self.last_egress_id_ui = last_egress_id_ui
        self.last_egress_id_mac = last_egress_id_mac
        self.next_egress_id_ui = next_egress_id_ui
        self.next_egress_id_mac = next_egress_id_mac

    @classmethod
    def parser(cls, buf):
        (type_, length,
         last_egress_id_ui, last_egress_id_mac,
         next_egress_id_ui, next_egress_id_mac
         ) = struct.unpack_from(cls._PACK_STR, buf)
        return cls(length,
                   last_egress_id_ui,
                   addrconv.mac.bin_to_text(last_egress_id_mac),
                   next_egress_id_ui,
                   addrconv.mac.bin_to_text(next_egress_id_mac))

    def serialize(self):
        # calculate length when it contains 0
        if self.length == 0:
            self.length = 16
        # start serialize
        buf = struct.pack(self._PACK_STR,
                          self._type,
                          self.length,
                          self.last_egress_id_ui,
                          addrconv.mac.text_to_bin(self.last_egress_id_mac),
                          self.next_egress_id_ui,
                          addrconv.mac.text_to_bin(self.next_egress_id_mac)
                          )
        return bytearray(buf)


@operation.register_tlv_types(CFM_ORGANIZATION_SPECIFIC_TLV)
class organization_specific_tlv(tlv):

    """CFM (IEEE Std 802.1ag-2007) Organization Specific TLV
       encoder/decoder class.

    This is used with os_ken.lib.packet.cfm.cfm.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ============== =======================================
    Attribute      Description
    ============== =======================================
    length         Length of Value field.
                   (0 means automatically-calculate when encoding.)
    oui            Organizationally Unique Identifier.
    subtype        Subtype.
    value          Value.(optional)
    ============== =======================================
    """

    _PACK_STR = '!BH3sB'
    _MIN_LEN = struct.calcsize(_PACK_STR)
    _OUI_AND_SUBTYPE_LEN = 4

    def __init__(self,
                 length=0,
                 oui=b"\x00\x00\x00",
                 subtype=0,
                 value=b""
                 ):
        super(organization_specific_tlv, self).__init__(length)
        self._type = CFM_ORGANIZATION_SPECIFIC_TLV
        self.oui = oui
        self.subtype = subtype
        self.value = value

    @classmethod
    def parser(cls, buf):
        (type_, length, oui, subtype) = struct.unpack_from(cls._PACK_STR, buf)
        value = b""
        if length > cls._OUI_AND_SUBTYPE_LEN:
            form = "%ds" % (length - cls._OUI_AND_SUBTYPE_LEN)
            (value,) = struct.unpack_from(form, buf, cls._MIN_LEN)
        return cls(length, oui, subtype, value)

    def serialize(self):
        # calculate length when it contains 0
        if self.length == 0:
            self.length = len(self.value) + self._OUI_AND_SUBTYPE_LEN
        # start serialize
        buf = struct.pack(self._PACK_STR,
                          self._type,
                          self.length,
                          self.oui,
                          self.subtype,
                          )
        buf = bytearray(buf)
        form = "%ds" % (self.length - self._OUI_AND_SUBTYPE_LEN)
        buf.extend(struct.pack(form, self.value))
        return buf


class reply_tlv(tlv):

    _PACK_STR = '!BHB6s'
    _MIN_LEN = struct.calcsize(_PACK_STR)
    _MIN_VALUE_LEN = _MIN_LEN - struct.calcsize('!BH')
    _PORT_ID_LENGTH_LEN = 1
    _PORT_ID_SUBTYPE_LEN = 1

    def __init__(self, length, action, mac_address, port_id_length,
                 port_id_subtype, port_id):
        super(reply_tlv, self).__init__(length)
        self.action = action
        self.mac_address = mac_address
        self.port_id_length = port_id_length
        self.port_id_subtype = port_id_subtype
        self.port_id = port_id

    @classmethod
    def parser(cls, buf):
        (type_, length, action,
         mac_address) = struct.unpack_from(cls._PACK_STR, buf)
        port_id_length = 0
        port_id_subtype = 0
        port_id = b''
        if length > cls._MIN_VALUE_LEN:
            (port_id_length,
             port_id_subtype) = struct.unpack_from('!2B', buf, cls._MIN_LEN)
            form = "%ds" % port_id_length
            (port_id,) = struct.unpack_from(form, buf,
                                            cls._MIN_LEN +
                                            cls._PORT_ID_LENGTH_LEN +
                                            cls._PORT_ID_SUBTYPE_LEN)
        return cls(length, action,
                   addrconv.mac.bin_to_text(mac_address),
                   port_id_length, port_id_subtype, port_id)

    def serialize(self):
        # calculate length when it contains 0
        if self.port_id_length == 0:
            self.port_id_length = len(self.port_id)
        if self.length == 0:
            self.length = self._MIN_VALUE_LEN
            if self.port_id_length != 0:
                self.length += (self.port_id_length +
                                self._PORT_ID_LENGTH_LEN +
                                self._PORT_ID_SUBTYPE_LEN)
        # start serialize
        buf = struct.pack(self._PACK_STR,
                          self._type,
                          self.length,
                          self.action,
                          addrconv.mac.text_to_bin(self.mac_address),
                          )
        buf = bytearray(buf)
        if self.port_id_length != 0:
            buf.extend(struct.pack("!BB",
                                   self.port_id_length,
                                   self.port_id_subtype))
            form = "%ds" % self.port_id_length
            buf.extend(struct.pack(form, self.port_id))
        return buf


@operation.register_tlv_types(CFM_REPLY_INGRESS_TLV)
class reply_ingress_tlv(reply_tlv):

    """CFM (IEEE Std 802.1ag-2007) Reply Ingress TLV encoder/decoder class.

    This is used with os_ken.lib.packet.cfm.cfm.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ================= =======================================
    Attribute         Description
    ================= =======================================
    length            Length of Value field.
                      (0 means automatically-calculate when encoding.)
    action            Ingress Action.The default is 1 (IngOK)
    mac_address       Ingress MAC Address.
    port_id_length    Ingress PortID Length.
                      (0 means automatically-calculate when encoding.)
    port_id_subtype   Ingress PortID Subtype.
    port_id           Ingress PortID.
    ================= =======================================
    """

    # Ingress Action field values
    _ING_OK = 1
    _ING_DOWN = 2
    _ING_BLOCKED = 3
    _ING_VID = 4

    def __init__(self,
                 length=0,
                 action=_ING_OK,
                 mac_address='00:00:00:00:00:00',
                 port_id_length=0,
                 port_id_subtype=0,
                 port_id=b''
                 ):
        super(reply_ingress_tlv, self).__init__(length, action,
                                                mac_address, port_id_length,
                                                port_id_subtype, port_id)
        assert action in [self._ING_OK, self._ING_DOWN,
                          self._ING_BLOCKED, self._ING_VID]
        self._type = CFM_REPLY_INGRESS_TLV


@operation.register_tlv_types(CFM_REPLY_EGRESS_TLV)
class reply_egress_tlv(reply_tlv):

    """CFM (IEEE Std 802.1ag-2007) Reply Egress TLV encoder/decoder class.

    This is used with os_ken.lib.packet.cfm.cfm.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    ================= =======================================
    Attribute         Description
    ================= =======================================
    length            Length of Value field.
                      (0 means automatically-calculate when encoding.)
    action            Egress Action.The default is 1 (EgrOK)
    mac_address       Egress MAC Address.
    port_id_length    Egress PortID Length.
                      (0 means automatically-calculate when encoding.)
    port_id_subtype   Egress PortID Subtype.
    port_id           Egress PortID.
    ================= =======================================
    """

    # Egress Action field values
    _EGR_OK = 1
    _EGR_DOWN = 2
    _EGR_BLOCKED = 3
    _EGR_VID = 4

    def __init__(self,
                 length=0,
                 action=_EGR_OK,
                 mac_address='00:00:00:00:00:00',
                 port_id_length=0,
                 port_id_subtype=0,
                 port_id=b''
                 ):
        super(reply_egress_tlv, self).__init__(length, action,
                                               mac_address, port_id_length,
                                               port_id_subtype, port_id)
        assert action in [self._EGR_OK, self._EGR_DOWN,
                          self._EGR_BLOCKED, self._EGR_VID]
        self._type = CFM_REPLY_EGRESS_TLV


operation.set_classes(operation._TLV_TYPES)



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from os_ken.lib.packet import packet_base


class itag(packet_base.PacketBase):
    """I-TAG (IEEE 802.1ah-2008) header encoder/decoder class.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    ============== ====================
    Attribute      Description
    ============== ====================
    pcp            Priority Code Point
    dei            Drop Eligible Indication
    uca            Use Customer Address
    sid            Service Instance ID
    ============== ====================
    """

    _PACK_STR = "!I"
    _MIN_LEN = struct.calcsize(_PACK_STR)

    def __init__(self, pcp=0, dei=0, uca=0, sid=0):
        super(itag, self).__init__()
        self.pcp = pcp
        self.dei = dei
        self.uca = uca
        self.sid = sid

    @classmethod
    def parser(cls, buf):
        (data, ) = struct.unpack_from(cls._PACK_STR, buf)
        pcp = data >> 29
        dei = data >> 28 & 1
        uca = data >> 27 & 1
        sid = data & 0x00ffffff
        # circular import: ethernet -> vlan -> pbb
        from os_ken.lib.packet import ethernet
        return (cls(pcp, dei, uca, sid), ethernet.ethernet,
                buf[cls._MIN_LEN:])

    def serialize(self, payload, prev):
        data = self.pcp << 29 | self.dei << 28 | self.uca << 27 | self.sid
        return struct.pack(self._PACK_STR, data)



# Copyright (C) 2017 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 Defines base data types and models required specifically
 for VRF Flow Specification support.
"""

import abc
import logging
import six

from os_ken.lib.packet.bgp import BGP_ATTR_TYPE_ORIGIN
from os_ken.lib.packet.bgp import BGP_ATTR_TYPE_AS_PATH
from os_ken.lib.packet.bgp import BGP_ATTR_TYPE_EXTENDED_COMMUNITIES
from os_ken.lib.packet.bgp import BGPPathAttributeOrigin
from os_ken.lib.packet.bgp import BGPPathAttributeAsPath
from os_ken.lib.packet.bgp import BGPPathAttributeExtendedCommunities

from os_ken.services.protocols.bgp.base import OrderedDict
from os_ken.services.protocols.bgp.info_base.vrf import VrfTable
from os_ken.services.protocols.bgp.info_base.vrf import VrfDest
from os_ken.services.protocols.bgp.info_base.vrf import VrfPath

from os_ken.services.protocols.bgp.utils.bgp import create_rt_extended_community

LOG = logging.getLogger('bgpspeaker.info_base.vrffs')


@six.add_metaclass(abc.ABCMeta)
class VRFFlowSpecTable(VrfTable):
    """Virtual Routing and Forwarding information base.
    Keeps destination imported to given VRF Flow Specification
    in represents.
    """

    def insert_vrffs_path(self, nlri, communities, is_withdraw=False):
        assert nlri
        assert isinstance(communities, list)
        vrf_conf = self.vrf_conf

        from os_ken.services.protocols.bgp.core import EXPECTED_ORIGIN
        pattrs = OrderedDict()
        pattrs[BGP_ATTR_TYPE_ORIGIN] = BGPPathAttributeOrigin(
            EXPECTED_ORIGIN)
        pattrs[BGP_ATTR_TYPE_AS_PATH] = BGPPathAttributeAsPath([])

        for rt in vrf_conf.export_rts:
            communities.append(create_rt_extended_community(rt, 2))
        for soo in vrf_conf.soo_list:
            communities.append(create_rt_extended_community(soo, 3))

        pattrs[BGP_ATTR_TYPE_EXTENDED_COMMUNITIES] = (
            BGPPathAttributeExtendedCommunities(communities=communities))

        puid = self.VRF_PATH_CLASS.create_puid(
            vrf_conf.route_dist, nlri.prefix)

        path = self.VRF_PATH_CLASS(
            puid, None, nlri, 0,
            pattrs=pattrs, is_withdraw=is_withdraw
        )

        # Insert the path into VRF table, get affected destination so that we
        # can process it further.
        eff_dest = self.insert(path)
        # Enqueue the eff_dest for further processing.
        self._signal_bus.dest_changed(eff_dest)


@six.add_metaclass(abc.ABCMeta)
class VRFFlowSpecDest(VrfDest):
    """Base class for VRF Flow Specification."""


@six.add_metaclass(abc.ABCMeta)
class VRFFlowSpecPath(VrfPath):
    """Represents a way of reaching an IP destination with
    a VPN Flow Specification.
    """



import logging
from time import strftime

from os_ken.services.protocols.bgp.operator.command import Command
from os_ken.services.protocols.bgp.operator.command import CommandsResponse
from os_ken.services.protocols.bgp.operator.command import STATUS_ERROR
from os_ken.services.protocols.bgp.operator.command import STATUS_OK
from os_ken.services.protocols.bgp.operator.commands.responses import \
    WrongParamResp
from os_ken.services.protocols.bgp.operator.views.bgp import CoreServiceDetailView
from os_ken.lib.packet.bgp import RF_IPv4_UC
from os_ken.lib.packet.bgp import RF_IPv6_UC
from os_ken.lib.packet.bgp import RF_IPv4_VPN
from os_ken.lib.packet.bgp import RF_IPv6_VPN
from os_ken.lib.packet.bgp import BGP_ATTR_ORIGIN_IGP
from os_ken.lib.packet.bgp import BGP_ATTR_ORIGIN_EGP
from os_ken.lib.packet.bgp import BGP_ATTR_ORIGIN_INCOMPLETE

LOG = logging.getLogger('bgpspeaker.operator.commands.show.neighbor')


class NeighborSummary(Command):
    help_msg = 'show summarized neighbor information'
    command = 'summary'

    def action(self, params):
        requested_peers = []
        if len(params) > 0:
            requested_peers = [str(p) for p in params]

        core_service = self.api.get_core_service()
        core_service_view = CoreServiceDetailView(core_service)
        peers_view = core_service_view.rel('peer_manager').rel('peers_summary')

        def filter_requested(peer_id, peer_obj):
            return not requested_peers or peer_id in requested_peers

        peers_view.apply_filter(filter_requested)
        ret = peers_view.encode()
        return CommandsResponse(STATUS_OK, ret)


class SentRoutes(Command):
    help_msg = 'paths sent and not withdrawn to given peer'
    command = 'sent-routes'
    param_help_msg = '<ip_addr> <addr_family>{vpnv4, vpnv6, ipv4, ipv6, all}'
    fmtstr = ' {0:<2s} {1:<19s} {2:<32s} {3:<8s} {4:<20s} '\
        '{5:<6s} {6:<6s} {7:<}\n'

    def action(self, params):
        if len(params) != 2:
            return WrongParamResp()
        ip_addr, addr_family = params

        if addr_family == 'ipv4':
            rf = RF_IPv4_UC
        elif addr_family == 'ipv6':
            rf = RF_IPv6_UC
        elif addr_family == 'vpnv4':
            rf = RF_IPv4_VPN
        elif addr_family == 'vpnv6':
            rf = RF_IPv6_VPN
        elif addr_family == 'all':
            rf = None
        else:
            return WrongParamResp('wrong addr_family name')

        ret = self._retrieve_paths(addr_family, rf, ip_addr).encode()
        return CommandsResponse(STATUS_OK, ret)

    def _retrieve_paths(self, addr_family, route_family, ip_addr):
        peer_view = self._retrieve_peer_view(ip_addr)
        adj_rib_out = peer_view.c_rel('adj_rib_out')
        adj_rib_out.apply_filter(lambda k, v: addr_family == 'all' or
                                 v.path.route_family == route_family)
        return adj_rib_out

    def _retrieve_peer_view(self, ip_addr):
        core_service = self.api.get_core_service()
        core_sv = CoreServiceDetailView(core_service)
        peers_view = core_sv.rel('peer_manager').rel('peers')
        peers_view.apply_filter(lambda k, v: v.ip_address == ip_addr)
        return peers_view

    @classmethod
    def cli_resp_formatter(cls, resp):
        if resp.status == STATUS_ERROR:
            return Command.cli_resp_formatter(resp)
        return cls._format_header() + cls._format_value(resp.value)

    @classmethod
    def _format_header(cls):
        ret = ''
        ret += ('Status codes: x filtered\n')
        ret += ('Origin codes: i - IGP, e - EGP, ? - incomplete\n')
        ret += cls.fmtstr.format('', 'Timestamp', 'Network', 'Labels',
                                 'Next Hop', 'Metric', 'LocPrf', 'Path')
        return ret

    @classmethod
    def _format_value(cls, value):
        ret = ''
        for v in value:
            path = v.get('path')
            aspath = path.get('as_path')
            origin = path.get('origin')

            if origin == BGP_ATTR_ORIGIN_IGP:
                origin = 'i'
            elif origin == BGP_ATTR_ORIGIN_EGP:
                origin = 'e'
            elif origin == BGP_ATTR_ORIGIN_INCOMPLETE:
                origin = '?'

            if origin:
                aspath = aspath + [origin]

            next_hop = path.get('nexthop')
            med = path.get('metric')
            labels = path.get('labels')
            localpref = path.get('local_pref')
            prefix = path.get('nlri').get('prefix')

            path_status = ''
            if v.get('filtered'):
                path_status = 'x'
            time = 'N/A'
            if v.get('timestamp'):
                time = strftime("%Y/%m/%d %H:%M:%S", v.get('timestamp'))
            ret += cls.fmtstr.format(path_status, time, prefix, labels,
                                     str(next_hop), str(med), str(localpref),
                                     ' '.join(map(str, aspath)))
        return ret


class ReceivedRoutes(SentRoutes):
    help_msg = 'paths received and not withdrawn by given peer'
    command = 'received-routes'

    def _retrieve_paths(self, addr_family, route_family, ip_addr):
        peer_view = self._retrieve_peer_view(ip_addr)
        adj_rib_in = peer_view.c_rel('adj_rib_in')
        adj_rib_in.apply_filter(lambda k, v: addr_family == 'all' or
                                v.path.route_family == route_family)
        return adj_rib_in


class Neighbor(Command):
    help_msg = 'show neighbor information'
    command = 'neighbor'
    subcommands = {
        'summary': NeighborSummary,
        'sent-routes': SentRoutes,
        'received-routes': ReceivedRoutes
    }

    fmtstr = ' {0:<12s} {1:<12s} {2:<}\n'

    def action(self, params):
        core_service = self.api.get_core_service()
        core_service_view = CoreServiceDetailView(core_service)
        peers_view = core_service_view.rel('peer_manager').rel('peers')

        ret = peers_view.encode()
        return CommandsResponse(STATUS_OK,
                                [{'ip_addr': k,
                                  'as_num': str(v['remote_as']),
                                  'bgp_state': v['stats']['bgp_state']}
                                 for k, v in ret.items()])

    @classmethod
    def cli_resp_formatter(cls, resp):
        if resp.status == STATUS_ERROR:
            return Command.cli_resp_formatter(resp)
        return cls._format_header() + cls._format_value(resp.value)

    @classmethod
    def _format_header(cls):
        return cls.fmtstr.format('IP Address', 'AS Number', 'BGP State')

    @classmethod
    def _format_value(cls, value):
        ret = ''
        for v in value:
            ret += cls.fmtstr.format(v['ip_addr'], v['as_num'], v['bgp_state'])
        return ret



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Internet Group Management Protocol(IGMP) packet parser/serializer

[RFC 1112] IGMP v1 format::

    0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |Version| Type  |    Unused     |           Checksum            |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                         Group Address                         |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

[RFC 2236] IGMP v2 format::

    0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |      Type     | Max Resp Time |           Checksum            |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                         Group Address                         |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

[RFC 3376] IGMP v3 Membership Query format::

    0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |  Type = 0x11  | Max Resp Code |           Checksum            |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                         Group Address                         |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   | Resv  |S| QRV |     QQIC      |     Number of Sources (N)     |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                       Source Address [1]                      |
   +-                                                             -+
   |                       Source Address [2]                      |
   +-                              .                              -+
   .                               .                               .
   .                               .                               .
   +-                                                             -+
   |                       Source Address [N]                      |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

IGMP v3 Membership Report format::

    0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |  Type = 0x22  |    Reserved   |           Checksum            |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |           Reserved            |  Number of Group Records (M)  |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                                                               |
   .                                                               .
   .                        Group Record [1]                       .
   .                                                               .
   |                                                               |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                                                               |
   .                                                               .
   .                        Group Record [2]                       .
   .                                                               .
   |                                                               |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                               .                               |
   .                               .                               .
   |                               .                               |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                                                               |
   .                                                               .
   .                        Group Record [M]                       .
   .                                                               .
   |                                                               |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

Where each Group Record has the following internal format::

   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |  Record Type  |  Aux Data Len |     Number of Sources (N)     |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                       Multicast Address                       |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                       Source Address [1]                      |
   +-                                                             -+
   |                       Source Address [2]                      |
   +-                                                             -+
   .                               .                               .
   .                               .                               .
   .                               .                               .
   +-                                                             -+
   |                       Source Address [N]                      |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                                                               |
   .                                                               .
   .                         Auxiliary Data                        .
   .                                                               .
   |                                                               |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
"""

import six
import struct
from math import trunc

from os_ken.lib import addrconv
from os_ken.lib import stringify
from os_ken.lib.packet import packet_base
from os_ken.lib.packet import packet_utils


IGMP_TYPE_QUERY = 0x11
IGMP_TYPE_REPORT_V1 = 0x12
IGMP_TYPE_REPORT_V2 = 0x16
IGMP_TYPE_LEAVE = 0x17
IGMP_TYPE_REPORT_V3 = 0x22

QUERY_RESPONSE_INTERVAL = 10.0
LAST_MEMBER_QUERY_INTERVAL = 1.0

MULTICAST_IP_ALL_HOST = '224.0.0.1'
MULTICAST_MAC_ALL_HOST = '01:00:5e:00:00:01'

# for types of IGMPv3 Report Group Records
MODE_IS_INCLUDE = 1
MODE_IS_EXCLUDE = 2
CHANGE_TO_INCLUDE_MODE = 3
CHANGE_TO_EXCLUDE_MODE = 4
ALLOW_NEW_SOURCES = 5
BLOCK_OLD_SOURCES = 6


class igmp(packet_base.PacketBase):
    """
    Internet Group Management Protocol(IGMP, RFC 1112, RFC 2236)
    header encoder/decoder class.

    http://www.ietf.org/rfc/rfc1112.txt

    http://www.ietf.org/rfc/rfc2236.txt

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte
    order.
    __init__ takes the corresponding args in this order.

    =============== ====================================================
    Attribute       Description
    =============== ====================================================
    msgtype         a message type for v2, or a combination of
                    version and a message type for v1.
    maxresp         max response time in unit of 1/10 second. it is
                    meaningful only in Query Message.
    csum            a check sum value. 0 means automatically-calculate
                    when encoding.
    address         a group address value.
    =============== ====================================================
    """
    _PACK_STR = '!BBH4s'
    _MIN_LEN = struct.calcsize(_PACK_STR)
    _TYPE = {
        'ascii': [
            'address'
        ]
    }

    def __init__(self, msgtype=IGMP_TYPE_QUERY, maxresp=0, csum=0,
                 address='0.0.0.0'):
        super(igmp, self).__init__()
        self.msgtype = msgtype
        self.maxresp = maxresp
        self.csum = csum
        self.address = address

    @classmethod
    def parser(cls, buf):
        assert cls._MIN_LEN <= len(buf)
        (msgtype, ) = struct.unpack_from('!B', buf)
        if (IGMP_TYPE_QUERY == msgtype and
                igmpv3_query.MIN_LEN <= len(buf)):
            (instance, subclass, rest,) = igmpv3_query.parser(buf)
        elif IGMP_TYPE_REPORT_V3 == msgtype:
            (instance, subclass, rest,) = igmpv3_report.parser(buf)
        else:
            (msgtype, maxresp, csum, address
             ) = struct.unpack_from(cls._PACK_STR, buf)
            instance = cls(msgtype, maxresp, csum,
                           addrconv.ipv4.bin_to_text(address))
            subclass = None
            rest = buf[cls._MIN_LEN:]
        return instance, subclass, rest

    def serialize(self, payload, prev):
        hdr = bytearray(struct.pack(self._PACK_STR, self.msgtype,
                                    trunc(self.maxresp), self.csum,
                                    addrconv.ipv4.text_to_bin(self.address)))

        if self.csum == 0:
            self.csum = packet_utils.checksum(hdr)
            struct.pack_into('!H', hdr, 2, self.csum)

        return hdr


class igmpv3_query(igmp):
    """
    Internet Group Management Protocol(IGMP, RFC 3376)
    Membership Query message encoder/decoder class.

    http://www.ietf.org/rfc/rfc3376.txt

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte
    order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    =============== ====================================================
    Attribute       Description
    =============== ====================================================
    msgtype         a message type for v3.
    maxresp         max response time in unit of 1/10 second.
    csum            a check sum value. 0 means automatically-calculate
                    when encoding.
    address         a group address value.
    s_flg           when set to 1, routers suppress the timer process.
    qrv             robustness variable for a querier.
    qqic            an interval time for a querier in unit of seconds.
    num             a number of the multicast servers.
    srcs            a list of IPv4 addresses of the multicast servers.
    =============== ====================================================
    """

    _PACK_STR = '!BBH4sBBH'
    _MIN_LEN = struct.calcsize(_PACK_STR)
    MIN_LEN = _MIN_LEN
    _TYPE = {
        'ascii': [
            'address'
        ],
        'asciilist': [
            'srcs'
        ]
    }

    def __init__(self, msgtype=IGMP_TYPE_QUERY, maxresp=100, csum=0,
                 address='0.0.0.0', s_flg=0, qrv=2, qqic=0, num=0,
                 srcs=None):
        super(igmpv3_query, self).__init__(
            msgtype, maxresp, csum, address)
        self.s_flg = s_flg
        self.qrv = qrv
        self.qqic = qqic
        self.num = num
        srcs = srcs or []
        assert isinstance(srcs, list)
        for src in srcs:
            assert isinstance(src, str)
        self.srcs = srcs

    @classmethod
    def parser(cls, buf):
        (msgtype, maxresp, csum, address, s_qrv, qqic, num
         ) = struct.unpack_from(cls._PACK_STR, buf)
        s_flg = (s_qrv >> 3) & 0b1
        qrv = s_qrv & 0b111
        offset = cls._MIN_LEN
        srcs = []
        while 0 < len(buf[offset:]) and num > len(srcs):
            assert 4 <= len(buf[offset:])
            (src, ) = struct.unpack_from('4s', buf, offset)
            srcs.append(addrconv.ipv4.bin_to_text(src))
            offset += 4
        assert num == len(srcs)
        return (cls(msgtype, maxresp, csum,
                    addrconv.ipv4.bin_to_text(address), s_flg, qrv,
                    qqic, num, srcs),
                None,
                buf[offset:])

    def serialize(self, payload, prev):
        s_qrv = self.s_flg << 3 | self.qrv
        buf = bytearray(struct.pack(self._PACK_STR, self.msgtype,
                                    trunc(self.maxresp), self.csum,
                                    addrconv.ipv4.text_to_bin(self.address),
                                    s_qrv, trunc(self.qqic), self.num))
        for src in self.srcs:
            buf.extend(struct.pack('4s', addrconv.ipv4.text_to_bin(src)))
        if 0 == self.num:
            self.num = len(self.srcs)
            struct.pack_into('!H', buf, 10, self.num)
        if 0 == self.csum:
            self.csum = packet_utils.checksum(buf)
            struct.pack_into('!H', buf, 2, self.csum)
        return six.binary_type(buf)

    def __len__(self):
        return self._MIN_LEN + len(self.srcs) * 4


class igmpv3_report(igmp):
    """
    Internet Group Management Protocol(IGMP, RFC 3376)
    Membership Report message encoder/decoder class.

    http://www.ietf.org/rfc/rfc3376.txt

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte
    order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    =============== ====================================================
    Attribute       Description
    =============== ====================================================
    msgtype         a message type for v3.
    csum            a check sum value. 0 means automatically-calculate
                    when encoding.
    record_num      a number of the group records.
    records         a list of os_ken.lib.packet.igmp.igmpv3_report_group.
                    None if no records.
    =============== ====================================================
    """

    _PACK_STR = '!BxH2xH'
    _MIN_LEN = struct.calcsize(_PACK_STR)
    _class_prefixes = ['igmpv3_report_group']

    def __init__(self, msgtype=IGMP_TYPE_REPORT_V3, csum=0, record_num=0,
                 records=None):
        self.msgtype = msgtype
        self.csum = csum
        self.record_num = record_num
        records = records or []
        assert isinstance(records, list)
        for record in records:
            assert isinstance(record, igmpv3_report_group)
        self.records = records

    @classmethod
    def parser(cls, buf):
        (msgtype, csum, record_num
         ) = struct.unpack_from(cls._PACK_STR, buf)
        offset = cls._MIN_LEN
        records = []
        while 0 < len(buf[offset:]) and record_num > len(records):
            record = igmpv3_report_group.parser(buf[offset:])
            records.append(record)
            offset += len(record)
        assert record_num == len(records)
        return (cls(msgtype, csum, record_num, records),
                None,
                buf[offset:])

    def serialize(self, payload, prev):
        buf = bytearray(struct.pack(self._PACK_STR, self.msgtype,
                                    self.csum, self.record_num))
        for record in self.records:
            buf.extend(record.serialize())
        if 0 == self.record_num:
            self.record_num = len(self.records)
            struct.pack_into('!H', buf, 6, self.record_num)
        if 0 == self.csum:
            self.csum = packet_utils.checksum(buf)
            struct.pack_into('!H', buf, 2, self.csum)
        return six.binary_type(buf)

    def __len__(self):
        records_len = 0
        for record in self.records:
            records_len += len(record)
        return self._MIN_LEN + records_len


class igmpv3_report_group(stringify.StringifyMixin):
    r"""
    Internet Group Management Protocol(IGMP, RFC 3376)
    Membership Report Group Record message encoder/decoder class.

    http://www.ietf.org/rfc/rfc3376.txt

    This is used with os_ken.lib.packet.igmp.igmpv3_report.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte
    order.
    __init__ takes the corresponding args in this order.

    .. tabularcolumns:: |l|L|

    =============== ====================================================
    Attribute       Description
    =============== ====================================================
    type\_          a group record type for v3.
    aux_len         the length of the auxiliary data.
    num             a number of the multicast servers.
    address         a group address value.
    srcs            a list of IPv4 addresses of the multicast servers.
    aux             the auxiliary data.
    =============== ====================================================
    """
    _PACK_STR = '!BBH4s'
    _MIN_LEN = struct.calcsize(_PACK_STR)
    _TYPE = {
        'ascii': [
            'address'
        ],
        'asciilist': [
            'srcs'
        ]
    }

    def __init__(self, type_=0, aux_len=0, num=0, address='0.0.0.0',
                 srcs=None, aux=None):
        self.type_ = type_
        self.aux_len = aux_len
        self.num = num
        self.address = address
        srcs = srcs or []
        assert isinstance(srcs, list)
        for src in srcs:
            assert isinstance(src, str)
        self.srcs = srcs
        self.aux = aux

    @classmethod
    def parser(cls, buf):
        (type_, aux_len, num, address
         ) = struct.unpack_from(cls._PACK_STR, buf)
        offset = cls._MIN_LEN
        srcs = []
        while 0 < len(buf[offset:]) and num > len(srcs):
            assert 4 <= len(buf[offset:])
            (src, ) = struct.unpack_from('4s', buf, offset)
            srcs.append(addrconv.ipv4.bin_to_text(src))
            offset += 4
        assert num == len(srcs)
        aux = None
        if aux_len:
            (aux, ) = struct.unpack_from('%ds' % (aux_len * 4), buf, offset)
        return cls(type_, aux_len, num,
                   addrconv.ipv4.bin_to_text(address), srcs, aux)

    def serialize(self):
        buf = bytearray(struct.pack(self._PACK_STR, self.type_,
                                    self.aux_len, self.num,
                                    addrconv.ipv4.text_to_bin(self.address)))
        for src in self.srcs:
            buf.extend(struct.pack('4s', addrconv.ipv4.text_to_bin(src)))
        if 0 == self.num:
            self.num = len(self.srcs)
            struct.pack_into('!H', buf, 2, self.num)
        if self.aux is not None:
            mod = len(self.aux) % 4
            if mod:
                self.aux += bytearray(4 - mod)
                self.aux = six.binary_type(self.aux)
            buf.extend(self.aux)
            if 0 == self.aux_len:
                self.aux_len = len(self.aux) // 4
                struct.pack_into('!B', buf, 1, self.aux_len)
        return six.binary_type(buf)

    def __len__(self):
        return self._MIN_LEN + len(self.srcs) * 4 + self.aux_len * 4



import logging
from collections import OrderedDict

import netaddr

from os_ken.services.protocols.bgp.base import SUPPORTED_GLOBAL_RF
from os_ken.services.protocols.bgp.info_base.rtc import RtcTable
from os_ken.services.protocols.bgp.info_base.ipv4 import Ipv4Path
from os_ken.services.protocols.bgp.info_base.ipv4 import Ipv4Table
from os_ken.services.protocols.bgp.info_base.ipv6 import Ipv6Path
from os_ken.services.protocols.bgp.info_base.ipv6 import Ipv6Table
from os_ken.services.protocols.bgp.info_base.vpnv4 import Vpnv4Table
from os_ken.services.protocols.bgp.info_base.vpnv6 import Vpnv6Table
from os_ken.services.protocols.bgp.info_base.vrf4 import Vrf4Table
from os_ken.services.protocols.bgp.info_base.vrf6 import Vrf6Table
from os_ken.services.protocols.bgp.info_base.vrfevpn import VrfEvpnTable
from os_ken.services.protocols.bgp.info_base.evpn import EvpnTable
from os_ken.services.protocols.bgp.info_base.ipv4fs import IPv4FlowSpecPath
from os_ken.services.protocols.bgp.info_base.ipv4fs import IPv4FlowSpecTable
from os_ken.services.protocols.bgp.info_base.vpnv4fs import VPNv4FlowSpecTable
from os_ken.services.protocols.bgp.info_base.vrf4fs import Vrf4FlowSpecTable
from os_ken.services.protocols.bgp.info_base.ipv6fs import IPv6FlowSpecPath
from os_ken.services.protocols.bgp.info_base.ipv6fs import IPv6FlowSpecTable
from os_ken.services.protocols.bgp.info_base.vpnv6fs import VPNv6FlowSpecTable
from os_ken.services.protocols.bgp.info_base.vrf6fs import Vrf6FlowSpecTable
from os_ken.services.protocols.bgp.info_base.l2vpnfs import L2VPNFlowSpecTable
from os_ken.services.protocols.bgp.info_base.vrfl2vpnfs import L2vpnFlowSpecPath
from os_ken.services.protocols.bgp.info_base.vrfl2vpnfs import L2vpnFlowSpecTable
from os_ken.services.protocols.bgp.rtconf.vrfs import VRF_RF_IPV4
from os_ken.services.protocols.bgp.rtconf.vrfs import VRF_RF_IPV6
from os_ken.services.protocols.bgp.rtconf.vrfs import VRF_RF_L2_EVPN
from os_ken.services.protocols.bgp.rtconf.vrfs import VRF_RF_IPV4_FLOWSPEC
from os_ken.services.protocols.bgp.rtconf.vrfs import VRF_RF_IPV6_FLOWSPEC
from os_ken.services.protocols.bgp.rtconf.vrfs import VRF_RF_L2VPN_FLOWSPEC
from os_ken.services.protocols.bgp.rtconf.vrfs import SUPPORTED_VRF_RF
from os_ken.services.protocols.bgp.utils.bgp import create_v4flowspec_actions
from os_ken.services.protocols.bgp.utils.bgp import create_v6flowspec_actions
from os_ken.services.protocols.bgp.utils.bgp import create_l2vpnflowspec_actions

from os_ken.lib import type_desc
from os_ken.lib import ip
from os_ken.lib.packet.bgp import RF_IPv4_UC
from os_ken.lib.packet.bgp import RF_IPv6_UC
from os_ken.lib.packet.bgp import RF_IPv4_VPN
from os_ken.lib.packet.bgp import RF_IPv6_VPN
from os_ken.lib.packet.bgp import RF_L2_EVPN
from os_ken.lib.packet.bgp import RF_IPv4_FLOWSPEC
from os_ken.lib.packet.bgp import RF_IPv6_FLOWSPEC
from os_ken.lib.packet.bgp import RF_VPNv4_FLOWSPEC
from os_ken.lib.packet.bgp import RF_VPNv6_FLOWSPEC
from os_ken.lib.packet.bgp import RF_L2VPN_FLOWSPEC
from os_ken.lib.packet.bgp import RF_RTC_UC
from os_ken.lib.packet.bgp import BGPPathAttributeOrigin
from os_ken.lib.packet.bgp import BGPPathAttributeAsPath
from os_ken.lib.packet.bgp import BGPPathAttributeExtendedCommunities
from os_ken.lib.packet.bgp import BGP_ATTR_TYPE_ORIGIN
from os_ken.lib.packet.bgp import BGP_ATTR_TYPE_AS_PATH
from os_ken.lib.packet.bgp import BGP_ATTR_ORIGIN_IGP
from os_ken.lib.packet.bgp import BGP_ATTR_TYPE_EXTENDED_COMMUNITIES
from os_ken.lib.packet.bgp import EvpnEsi
from os_ken.lib.packet.bgp import EvpnArbitraryEsi
from os_ken.lib.packet.bgp import EvpnNLRI
from os_ken.lib.packet.bgp import EvpnMacIPAdvertisementNLRI
from os_ken.lib.packet.bgp import EvpnInclusiveMulticastEthernetTagNLRI
from os_ken.lib.packet.bgp import IPAddrPrefix
from os_ken.lib.packet.bgp import IP6AddrPrefix
from os_ken.lib.packet.bgp import FlowSpecIPv4NLRI
from os_ken.lib.packet.bgp import FlowSpecIPv6NLRI
from os_ken.lib.packet.bgp import FlowSpecL2VPNNLRI

from os_ken.services.protocols.bgp.utils.validation import is_valid_ipv4
from os_ken.services.protocols.bgp.utils.validation import is_valid_ipv4_prefix
from os_ken.services.protocols.bgp.utils.validation import is_valid_ipv6
from os_ken.services.protocols.bgp.utils.validation import is_valid_ipv6_prefix


LOG = logging.getLogger('bgpspeaker.core_managers.table_mixin')


class TableCoreManager(object):
    """Methods performing core operations on tables."""

    def __init__(self, core_service, common_conf):

        self._tables = {}
        self._rt_mgr = core_service.rt_manager
        self._signal_bus = core_service.signal_bus

        # (VRF) Tables to which the routes with a given route target
        # should be imported.
        #
        # Key: RouteTarget
        # Value: List of tables.
        self._tables_for_rt = {}

        # Global/Default tables, keyed by RouteFamily.
        self._global_tables = {}

        self._core_service = core_service
        self._signal_bus = self._core_service.signal_bus

        # VPN label range
        self._asbr_label_range = common_conf.label_range

        self._next_vpnv4_label = int(self._asbr_label_range[0])

        self._next_hop_label = {}

    @property
    def global_tables(self):
        return self._global_tables

    def remove_vrf_by_vrf_conf(self, vrf_conf):

        route_family = vrf_conf.route_family
        assert route_family in SUPPORTED_VRF_RF
        table_id = (vrf_conf.route_dist, route_family)

        vrf_table = self._tables.pop(table_id)

        self._remove_links_to_vrf_table(vrf_table)

        # Withdraw the best-path whose source was NC since it may have been
        # exported to VPN table.
        for destination in vrf_table.values():
            best_path = destination.best_path
            if best_path and best_path.source is None:
                vpn_clone = best_path.clone_to_vpn(vrf_conf.route_dist,
                                                   for_withdrawal=True)
                self.learn_path(vpn_clone)
        LOG.debug('VRF with RD %s marked for removal', vrf_conf.route_dist)

    def import_all_vpn_paths_to_vrf(self, vrf_table, import_rts=None):
        """Imports VPNv4/6 or EVPN paths from Global/VPN table into given
        VRFv4/6  or VRFEVPN table.
        :param vrf_table: Vrf table to which we import
        :type vrf_table: VrfTable
        :param import_rts: import RTs to override default import_rts of
         vrf table for this import
        :type import_rts: set of strings

        Checks if we have any path RT common with VRF table's import RT.
        """
        if vrf_table.route_family == Vrf4Table.ROUTE_FAMILY:
            vpn_table = self.get_vpn4_table()
        elif vrf_table.route_family == Vrf6Table.ROUTE_FAMILY:
            vpn_table = self.get_vpn6_table()
        elif vrf_table.route_family == VrfEvpnTable.ROUTE_FAMILY:
            vpn_table = self.get_evpn_table()
        elif vrf_table.route_family == Vrf4FlowSpecTable.ROUTE_FAMILY:
            vpn_table = self.get_vpnv4fs_table()
        elif vrf_table.route_family == Vrf6FlowSpecTable.ROUTE_FAMILY:
            vpn_table = self.get_vpnv6fs_table()
        elif vrf_table.route_family == L2vpnFlowSpecTable.ROUTE_FAMILY:
            vpn_table = self.get_l2vpnfs_table()
        else:
            raise ValueError('Invalid VRF table route family: %s' %
                             vrf_table.route_family)

        vrf_table.import_vpn_paths_from_table(vpn_table, import_rts)

    def learn_path(self, path):
        """Inserts `path` into correct global table.

        Since known paths to `Destination` has changes, we queue it for further
        processing.
        """
        # Get VPN/Global table
        table = self.get_global_table_by_route_family(path.route_family)
        gpath_dest = table.insert(path)
        # Since destination was updated, we enqueue it for processing.
        self._signal_bus.dest_changed(gpath_dest)

    def remember_sent_route(self, sent_route):
        """Records `sent_route` inside proper table.

        Records of `sent_route` from Adj-RIB-out.
        """
        route_family = sent_route.path.route_family
        table = self.get_global_table_by_route_family(route_family)
        table.insert_sent_route(sent_route)

    def on_interesting_rts_change(self, new_global_rts, removed_global_rts):
        """Update global tables as interested RTs changed.

        Adds `new_rts` and removes `removed_rts` rt nlris. Does not check if
        `new_rts` or `removed_rts` are already present. Schedules refresh
        request to peers that do not participate in RTC address-family.
        """
        # We add new RT NLRI and request RR for other peers.
        if new_global_rts:
            LOG.debug(
                'Sending route_refresh to all neighbors that'
                ' did not negotiate RTC capability.'
            )

            pm = self._core_service.peer_manager
            pm.schedule_rr_to_non_rtc_peers()
        if removed_global_rts:
            LOG.debug(
                'Cleaning up global tables as some interested RTs were removed'
            )
            self._clean_global_uninteresting_paths()

    def get_global_table_by_route_family(self, route_family):
        if route_family not in SUPPORTED_GLOBAL_RF:
            raise ValueError(
                'Given route family: %s currently not supported' % route_family
            )

        global_table = None
        if route_family == RF_IPv4_UC:
            global_table = self.get_ipv4_table()
        elif route_family == RF_IPv6_UC:
            global_table = self.get_ipv6_table()
        elif route_family == RF_IPv4_VPN:
            global_table = self.get_vpn4_table()
        elif route_family == RF_IPv6_VPN:
            global_table = self.get_vpn6_table()
        elif route_family == RF_L2_EVPN:
            global_table = self.get_evpn_table()
        elif route_family == RF_IPv4_FLOWSPEC:
            global_table = self.get_ipv4fs_table()
        elif route_family == RF_IPv6_FLOWSPEC:
            global_table = self.get_ipv6fs_table()
        elif route_family == RF_VPNv4_FLOWSPEC:
            global_table = self.get_vpnv4fs_table()
        elif route_family == RF_VPNv6_FLOWSPEC:
            global_table = self.get_vpnv6fs_table()
        elif route_family == RF_L2VPN_FLOWSPEC:
            global_table = self.get_l2vpnfs_table()
        elif route_family == RF_RTC_UC:
            global_table = self.get_rtc_table()

        return global_table

    def get_vrf_table(self, vrf_rd, vrf_rf):
        assert vrf_rd is not None
        return self._tables.get((vrf_rd, vrf_rf))

    def get_vrf_tables(self, vrf_rf=None):
        vrf_tables = {}
        for (scope_id, table_id), table in self._tables.items():
            if scope_id is None:
                continue
            if vrf_rf is not None and table_id != vrf_rf:
                continue
            vrf_tables[(scope_id, table_id)] = table
        return vrf_tables

    def get_ipv4_table(self):
        """Returns global IPv4 table.

        Creates the table if it does not exist.
        """

        vpn_table = self._global_tables.get(RF_IPv4_UC)
        # Lazy initialize the table.
        if not vpn_table:
            vpn_table = Ipv4Table(self._core_service, self._signal_bus)
            self._global_tables[RF_IPv4_UC] = vpn_table
            self._tables[(None, RF_IPv4_UC)] = vpn_table

        return vpn_table

    def get_ipv6_table(self):
        table = self._global_tables.get(RF_IPv6_UC)
        if not table:
            table = Ipv6Table(self._core_service, self._signal_bus)
            self._global_tables[RF_IPv6_UC] = table
            self._tables[(None, RF_IPv6_UC)] = table
        return table

    def get_vpn6_table(self):
        """Returns global VPNv6 table.

        Creates the table if it does not exist.
        """
        vpn_table = self._global_tables.get(RF_IPv6_VPN)
        # Lazy initialize the table.
        if not vpn_table:
            vpn_table = Vpnv6Table(self._core_service, self._signal_bus)
            self._global_tables[RF_IPv6_VPN] = vpn_table
            self._tables[(None, RF_IPv6_VPN)] = vpn_table

        return vpn_table

    def get_vpn4_table(self):
        """Returns global VPNv6 table.

        Creates the table if it does not exist.
        """
        vpn_table = self._global_tables.get(RF_IPv4_VPN)
        # Lazy initialize the table.
        if not vpn_table:
            vpn_table = Vpnv4Table(self._core_service, self._signal_bus)
            self._global_tables[RF_IPv4_VPN] = vpn_table
            self._tables[(None, RF_IPv4_VPN)] = vpn_table

        return vpn_table

    def get_evpn_table(self):
        """Returns global EVPN table.

        Creates the table if it does not exist.
        """
        evpn_table = self._global_tables.get(RF_L2_EVPN)
        # Lazy initialization of the table.
        if not evpn_table:
            evpn_table = EvpnTable(self._core_service, self._signal_bus)
            self._global_tables[RF_L2_EVPN] = evpn_table
            self._tables[(None, RF_L2_EVPN)] = evpn_table

        return evpn_table

    def get_rtc_table(self):
        """Returns global RTC table.

        Creates the table if it does not exist.
        """
        rtc_table = self._global_tables.get(RF_RTC_UC)
        # Lazy initialization of the table.
        if not rtc_table:
            rtc_table = RtcTable(self._core_service, self._signal_bus)
            self._global_tables[RF_RTC_UC] = rtc_table
            self._tables[(None, RF_RTC_UC)] = rtc_table
        return rtc_table

    def get_next_vpnv4_label(self):
        # Get next available label
        lbl = self._next_vpnv4_label
        # Check if label is within max. range allowed.
        if lbl > int(self._asbr_label_range[1]):
            # Currently we log error message if we exceed configured range.
            message = 'Have reached max label range'
            LOG.error(message)
            raise ValueError(message)
            # Increment label by 1 as next label.
        self._next_vpnv4_label += 1
        return lbl

    def get_ipv4fs_table(self):
        """Returns global IPv4 Flow Specification table.

        Creates the table if it does not exist.
        """
        ipv4fs_table = self._global_tables.get(RF_IPv4_FLOWSPEC)
        # Lazy initialization of the table.
        if not ipv4fs_table:
            ipv4fs_table = IPv4FlowSpecTable(self._core_service,
                                             self._signal_bus)
            self._global_tables[RF_IPv4_FLOWSPEC] = ipv4fs_table
            self._tables[(None, RF_IPv4_FLOWSPEC)] = ipv4fs_table

        return ipv4fs_table

    def get_ipv6fs_table(self):
        """Returns global IPv6 Flow Specification table.

        Creates the table if it does not exist.
        """
        ipv6fs_table = self._global_tables.get(RF_IPv6_FLOWSPEC)
        # Lazy initialization of the table.
        if not ipv6fs_table:
            ipv6fs_table = IPv6FlowSpecTable(self._core_service,
                                             self._signal_bus)
            self._global_tables[RF_IPv6_FLOWSPEC] = ipv6fs_table
            self._tables[(None, RF_IPv6_FLOWSPEC)] = ipv6fs_table

        return ipv6fs_table

    def get_vpnv4fs_table(self):
        """Returns global VPNv4 Flow Specification table.

        Creates the table if it does not exist.
        """
        vpnv4fs_table = self._global_tables.get(RF_VPNv4_FLOWSPEC)
        # Lazy initialization of the table.
        if not vpnv4fs_table:
            vpnv4fs_table = VPNv4FlowSpecTable(self._core_service,
                                               self._signal_bus)
            self._global_tables[RF_VPNv4_FLOWSPEC] = vpnv4fs_table
            self._tables[(None, RF_VPNv4_FLOWSPEC)] = vpnv4fs_table

        return vpnv4fs_table

    def get_vpnv6fs_table(self):
        """Returns global VPNv6 Flow Specification table.

        Creates the table if it does not exist.
        """
        vpnv6fs_table = self._global_tables.get(RF_VPNv6_FLOWSPEC)
        # Lazy initialization of the table.
        if not vpnv6fs_table:
            vpnv6fs_table = VPNv6FlowSpecTable(self._core_service,
                                               self._signal_bus)
            self._global_tables[RF_VPNv6_FLOWSPEC] = vpnv6fs_table
            self._tables[(None, RF_VPNv6_FLOWSPEC)] = vpnv6fs_table

        return vpnv6fs_table

    def get_l2vpnfs_table(self):
        """Returns global L2VPN Flow Specification table.

        Creates the table if it does not exist.
        """
        l2vpnfs_table = self._global_tables.get(RF_L2VPN_FLOWSPEC)
        # Lazy initialization of the table.
        if not l2vpnfs_table:
            l2vpnfs_table = L2VPNFlowSpecTable(self._core_service,
                                               self._signal_bus)
            self._global_tables[RF_L2VPN_FLOWSPEC] = l2vpnfs_table
            self._tables[(None, RF_L2VPN_FLOWSPEC)] = l2vpnfs_table

        return l2vpnfs_table

    def get_nexthop_label(self, label_key):
        return self._next_hop_label.get(label_key, None)

    def set_nexthop_label(self, key, value):
        self._next_hop_label[key] = value

    def update_vrf_table_links(self, vrf_table, new_imp_rts,
                               removed_imp_rts):
        """Update mapping from RT to VRF table."""
        assert vrf_table
        if new_imp_rts:
            self._link_vrf_table(vrf_table, new_imp_rts)
        if removed_imp_rts:
            self._remove_links_to_vrf_table_for_rts(vrf_table,
                                                    removed_imp_rts)

    def re_install_net_ctrl_paths(self, vrf_table):
        """Re-installs paths from NC with current BGP policy.

        Iterates over known paths from NC installed in `vrf4_table` and
        adds new path with path attributes as per current VRF configuration.
        """
        assert vrf_table
        for dest in vrf_table.values():
            for path in dest.known_path_list:
                if path.source is None:
                    vrf_table.insert_vrf_path(
                        nlri=path.nlri,
                        next_hop=path.nexthop,
                        gen_lbl=True
                    )
        LOG.debug('Re-installed NC paths with current policy for table %s.',
                  vrf_table)

    def _remove_links_to_vrf_table(self, vrf_table):
        """Removes any links to given `vrf_table`."""
        assert vrf_table
        vrf_conf = vrf_table.vrf_conf
        self._remove_links_to_vrf_table_for_rts(vrf_table,
                                                vrf_conf.import_rts)

    def _remove_links_to_vrf_table_for_rts(self, vrf_table, rts):
        rts_with_no_table = set()
        affected_tables = set()
        route_family = vrf_table.route_family
        for rt in rts:
            rt_rf_id = rt + ':' + str(route_family)
            rt_specific_tables = self._tables_for_rt.get(rt_rf_id)
            affected_tables.update(rt_specific_tables)
            if rt_specific_tables:
                try:
                    rt_specific_tables.remove(vrf_table)
                except KeyError:
                    LOG.debug('Did not find table listed as interested '
                              'for its import RT: %s', rt)
                if len(rt_specific_tables) == 0:
                    rts_with_no_table.add(rt)

        # Remove records of RT that have no tables associated with it.
        for rt in rts_with_no_table:
            rt_rf_id = rt + ':' + str(route_family)
            del self._tables_for_rt[rt_rf_id]

    def create_and_link_vrf_table(self, vrf_conf):
        """Factory method to create VRF table for given `vrf_conf`.

        Adds mapping to this table with appropriate scope. Also, adds mapping
        for import RT of this VRF to created table to facilitate
        importing/installing of paths from global tables.
        Returns created table.
        """
        route_family = vrf_conf.route_family

        if route_family == VRF_RF_IPV4:
            vrf_table = Vrf4Table
        elif route_family == VRF_RF_IPV6:
            vrf_table = Vrf6Table
        elif route_family == VRF_RF_L2_EVPN:
            vrf_table = VrfEvpnTable
        elif route_family == VRF_RF_IPV4_FLOWSPEC:
            vrf_table = Vrf4FlowSpecTable
        elif route_family == VRF_RF_IPV6_FLOWSPEC:
            vrf_table = Vrf6FlowSpecTable
        elif route_family == VRF_RF_L2VPN_FLOWSPEC:
            vrf_table = L2vpnFlowSpecTable
        else:
            raise ValueError('Unsupported route family for VRF: %s' %
                             route_family)

        vrf_table = vrf_table(vrf_conf, self._core_service, self._signal_bus)
        table_id = (vrf_conf.route_dist, route_family)
        self._tables[table_id] = vrf_table

        assert vrf_table is not None
        LOG.debug('Added new VrfTable with route_dist:%s and route_family:%s',
                  vrf_conf.route_dist, route_family)

        import_rts = vrf_conf.import_rts
        # If VRF is configured with import RT, we put this table
        # in a list corresponding to this RT for easy access.
        if import_rts:
            self._link_vrf_table(vrf_table, import_rts)

        return vrf_table

    def _link_vrf_table(self, vrf_table, rt_list):
        route_family = vrf_table.route_family
        for rt in rt_list:
            rt_rf_id = rt + ':' + str(route_family)
            table_set = self._tables_for_rt.get(rt_rf_id)
            if table_set is None:
                table_set = set()
                self._tables_for_rt[rt_rf_id] = table_set
            table_set.add(vrf_table)
            LOG.debug('Added VrfTable %s to import RT table list: %s',
                      vrf_table, rt)

    def _clean_global_uninteresting_paths(self):
        """Marks paths that do not have any route targets of interest
        for withdrawal.

        Since global tables can have paths with route targets that are not
        interesting any more, we have to clean these paths so that appropriate
        withdraw are sent out to NC and other peers. Interesting route targets
        change as VRF are modified or some filter is that specify what route
        targets are allowed are updated. This clean up should only be done when
        a route target is no longer considered interesting and some paths with
        that route target was installing in any of the global table.
        """
        uninteresting_dest_count = 0
        interested_rts = self._rt_mgr.global_interested_rts
        LOG.debug('Cleaning uninteresting paths. Global interested RTs %s',
                  interested_rts)
        for route_family in [RF_IPv4_VPN, RF_IPv6_VPN, RF_RTC_UC]:
            # TODO(PH): We currently do not install RT_NLRI paths based on
            # extended path attributes (RT)
            if route_family == RF_RTC_UC:
                continue
            table = self.get_global_table_by_route_family(route_family)
            uninteresting_dest_count += \
                table.clean_uninteresting_paths(interested_rts)

        LOG.debug('Found %s number of destinations had uninteresting paths.',
                  uninteresting_dest_count)

    def import_single_vpn_path_to_all_vrfs(self, vpn_path, path_rts=None):
        """Imports *vpn_path* to qualifying VRF tables.

        Import RTs of VRF table is matched with RTs from *vpn4_path* and if we
        have any common RTs we import the path into VRF.
        """
        LOG.debug('Importing path %s to qualifying VRFs', vpn_path)

        # If this path has no RTs we are done.
        if not path_rts:
            LOG.info('Encountered a path with no RTs: %s', vpn_path)
            return

        # We match path RTs with all VRFs that are interested in them.
        interested_tables = set()

        # Get route family of VRF to when this VPN Path can be imported to
        if vpn_path.route_family == RF_IPv4_VPN:
            route_family = RF_IPv4_UC
        elif vpn_path.route_family == RF_IPv6_VPN:
            route_family = RF_IPv6_UC
        elif vpn_path.route_family == RF_L2_EVPN:
            route_family = RF_L2_EVPN
        elif vpn_path.route_family == RF_VPNv4_FLOWSPEC:
            route_family = RF_IPv4_FLOWSPEC
        elif vpn_path.route_family == RF_VPNv6_FLOWSPEC:
            route_family = RF_IPv6_FLOWSPEC
        elif vpn_path.route_family == RF_L2VPN_FLOWSPEC:
            route_family = RF_L2VPN_FLOWSPEC
        else:
            raise ValueError('Unsupported route family for VRF: %s' %
                             vpn_path.route_family)

        for rt in path_rts:
            rt_rf_id = rt + ':' + str(route_family)
            vrf_rt_tables = self._tables_for_rt.get(rt_rf_id)
            if vrf_rt_tables:
                interested_tables.update(vrf_rt_tables)

        if interested_tables:
            # We iterate over all VRF tables that are interested in the RT
            # of the given path and import this path into them.
            route_dist = vpn_path.nlri.route_dist
            for vrf_table in interested_tables:
                if (vpn_path.source is not None or
                        route_dist != vrf_table.vrf_conf.route_dist):
                    update_vrf_dest = vrf_table.import_vpn_path(vpn_path)
                    # Queue the destination for further processing.
                    if update_vrf_dest is not None:
                        self._signal_bus.\
                            dest_changed(update_vrf_dest)
        else:
            # If we do not have any VRF with import RT that match with path RT
            LOG.debug('No VRF table found that imports RTs: %s', path_rts)

    def update_vrf_table(self, route_dist, prefix=None, next_hop=None,
                         route_family=None, route_type=None, tunnel_type=None,
                         is_withdraw=False, redundancy_mode=None,
                         pmsi_tunnel_type=None, tunnel_endpoint_ip=None,
                         mac_mobility=None, **kwargs):
        """Update a BGP route in the VRF table identified by `route_dist`
        with the given `next_hop`.

        If `is_withdraw` is False, which is the default, add a BGP route
        to the VRF table identified by `route_dist` with the given
        `next_hop`.
        If `is_withdraw` is True, remove a BGP route from the VRF table
        and the given `next_hop` is ignored.

        If `route_family` is VRF_RF_L2_EVPN, `route_type` and `kwargs`
        are required to construct EVPN NLRI and `prefix` is ignored.

        ``redundancy_mode`` specifies a redundancy mode type.

`       `pmsi_tunnel_type` specifies the type of the PMSI tunnel attribute
         used to encode the multicast tunnel identifier.
        This field is advertised only if route_type is
        EVPN_MULTICAST_ETAG_ROUTE.

        `tunnel_endpoint_ip` specifies a tunnel endpoint IP other than the
        default local router ID; only used in EVPN_MULTICAST_ETAG_ROUTE

        `mac_mobility` specifies an optional integer sequence number to insert
        as a MAC Mobility extended community; special value `-1` is used for
        static MACs (MAC Mobility sequence 0 with STATIC flag set)

        Returns assigned VPN label.
        """
        from os_ken.services.protocols.bgp.core import BgpCoreError

        assert route_dist

        if is_withdraw:
            gen_lbl = False
            next_hop = None
        else:
            gen_lbl = True
            if not (is_valid_ipv4(next_hop) or is_valid_ipv6(next_hop)):
                raise BgpCoreError(
                    desc='Invalid IPv4/IPv6 nexthop: %s' % next_hop)

        vrf_table = self._tables.get((route_dist, route_family))
        if vrf_table is None:
            raise BgpCoreError(
                desc='VRF table  does not exist: route_dist=%s, '
                     'route_family=%s' % (route_dist, route_family))

        vni = kwargs.get('vni', None)

        if route_family == VRF_RF_IPV4:
            if not is_valid_ipv4_prefix(prefix):
                raise BgpCoreError(desc='Invalid IPv4 prefix: %s' % prefix)
            ip, masklen = prefix.split('/')
            prefix = IPAddrPrefix(int(masklen), ip)
        elif route_family == VRF_RF_IPV6:
            if not is_valid_ipv6_prefix(prefix):
                raise BgpCoreError(desc='Invalid IPv6 prefix: %s' % prefix)
            ip6, masklen = prefix.split('/')
            prefix = IP6AddrPrefix(int(masklen), ip6)
        elif route_family == VRF_RF_L2_EVPN:
            assert route_type
            if route_type == EvpnMacIPAdvertisementNLRI.ROUTE_TYPE_NAME:
                # MPLS labels will be assigned automatically
                kwargs['mpls_labels'] = []
            if route_type == EvpnInclusiveMulticastEthernetTagNLRI.ROUTE_TYPE_NAME:
                # Inclusive Multicast Ethernet Tag Route does not have "vni",
                # omit "vni" from "kwargs" here.
                vni = kwargs.pop('vni', None)
            subclass = EvpnNLRI._lookup_type_name(route_type)
            kwargs['route_dist'] = route_dist
            esi = kwargs.get('esi', None)
            if esi is not None:
                if isinstance(esi, dict):
                    esi_type = esi.get('type', 0)
                    esi_class = EvpnEsi._lookup_type(esi_type)
                    kwargs['esi'] = esi_class.from_jsondict(esi)
                else:  # isinstance(esi, numbers.Integral)
                    kwargs['esi'] = EvpnArbitraryEsi(
                        type_desc.Int9.from_user(esi))
            if vni is not None:
                # Disable to generate MPLS labels,
                # because encapsulation type is not MPLS.
                from os_ken.services.protocols.bgp.api.prefix import (
                    TUNNEL_TYPE_VXLAN, TUNNEL_TYPE_NVGRE)
                assert tunnel_type in [
                    None, TUNNEL_TYPE_VXLAN, TUNNEL_TYPE_NVGRE]
                gen_lbl = False
            prefix = subclass(**kwargs)
        else:
            raise BgpCoreError(
                desc='Unsupported route family %s' % route_family)

        # We do not check if we have a path to given prefix, we issue
        # withdrawal. Hence multiple withdrawals have not side effect.
        return vrf_table.insert_vrf_path(
            nlri=prefix, next_hop=next_hop, gen_lbl=gen_lbl,
            is_withdraw=is_withdraw, redundancy_mode=redundancy_mode,
            mac_mobility=mac_mobility,
            vni=vni, tunnel_type=tunnel_type,
            pmsi_tunnel_type=pmsi_tunnel_type)

    def update_flowspec_vrf_table(self, flowspec_family, route_dist, rules,
                                  actions=None, is_withdraw=False):
        """Update a BGP route in the VRF table for Flow Specification.

        ``flowspec_family`` specifies one of the flowspec family name.

        ``route_dist`` specifies a route distinguisher value.

        ``rules`` specifies NLRIs of Flow Specification as
        a dictionary type value.

        `` actions`` specifies Traffic Filtering Actions of
        Flow Specification as a dictionary type value.

        If `is_withdraw` is False, which is the default, add a BGP route
        to the VRF table identified by `route_dist`.
        If `is_withdraw` is True, remove a BGP route from the VRF table.
        """
        from os_ken.services.protocols.bgp.core import BgpCoreError
        from os_ken.services.protocols.bgp.api.prefix import (
            FLOWSPEC_FAMILY_VPNV4,
            FLOWSPEC_FAMILY_VPNV6,
            FLOWSPEC_FAMILY_L2VPN,
        )

        if flowspec_family == FLOWSPEC_FAMILY_VPNV4:
            vrf_table = self._tables.get((route_dist, VRF_RF_IPV4_FLOWSPEC))
            prefix = FlowSpecIPv4NLRI.from_user(**rules)
            try:
                communities = create_v4flowspec_actions(actions)
            except ValueError as e:
                raise BgpCoreError(desc=str(e))
        elif flowspec_family == FLOWSPEC_FAMILY_VPNV6:
            vrf_table = self._tables.get((route_dist, VRF_RF_IPV6_FLOWSPEC))
            prefix = FlowSpecIPv6NLRI.from_user(**rules)
            try:
                communities = create_v6flowspec_actions(actions)
            except ValueError as e:
                raise BgpCoreError(desc=str(e))
        elif flowspec_family == FLOWSPEC_FAMILY_L2VPN:
            vrf_table = self._tables.get((route_dist, VRF_RF_L2VPN_FLOWSPEC))
            prefix = FlowSpecL2VPNNLRI.from_user(route_dist, **rules)
            try:
                communities = create_l2vpnflowspec_actions(actions)
            except ValueError as e:
                raise BgpCoreError(desc=str(e))
        else:
            raise BgpCoreError(
                desc='Unsupported flowspec_family %s' % flowspec_family)

        if vrf_table is None:
            raise BgpCoreError(
                desc='VRF table does not exist: route_dist=%s, '
                     'flowspec_family=%s' % (route_dist, flowspec_family))

        # We do not check if we have a path to given prefix, we issue
        # withdrawal. Hence multiple withdrawals have not side effect.
        vrf_table.insert_vrffs_path(
            nlri=prefix, communities=communities,
            is_withdraw=is_withdraw)

    def update_global_table(self, prefix, next_hop=None, is_withdraw=False):
        """Update a BGP route in the Global table for the given `prefix`
        with the given `next_hop`.

        If `is_withdraw` is False, which is the default, add a BGP route
        to the Global table.
        If `is_withdraw` is True, remove a BGP route from the Global table.
        """
        src_ver_num = 1
        peer = None
        # set mandatory path attributes
        origin = BGPPathAttributeOrigin(BGP_ATTR_ORIGIN_IGP)
        aspath = BGPPathAttributeAsPath([[]])

        pathattrs = OrderedDict()
        pathattrs[BGP_ATTR_TYPE_ORIGIN] = origin
        pathattrs[BGP_ATTR_TYPE_AS_PATH] = aspath

        net = netaddr.IPNetwork(prefix)
        addr = str(net.ip)
        masklen = net.prefixlen
        if ip.valid_ipv4(addr):
            _nlri = IPAddrPrefix(masklen, addr)
            if next_hop is None:
                next_hop = '0.0.0.0'
            p = Ipv4Path
        else:
            _nlri = IP6AddrPrefix(masklen, addr)
            if next_hop is None:
                next_hop = '::'
            p = Ipv6Path

        new_path = p(peer, _nlri, src_ver_num,
                     pattrs=pathattrs, nexthop=next_hop,
                     is_withdraw=is_withdraw)

        # add to global table and propagates to neighbors
        self.learn_path(new_path)

    def update_flowspec_global_table(self, flowspec_family, rules,
                                     actions=None, is_withdraw=False):
        """Update a BGP route in the Global table for Flow Specification.

        ``flowspec_family`` specifies one of the Flow Specification
         family name.

        ``rules`` specifies NLRIs of Flow Specification as
        a dictionary type value.

        `` actions`` specifies Traffic Filtering Actions of
        Flow Specification as a dictionary type value.

        If `is_withdraw` is False, which is the default, add a BGP route
        to the Global table.
        If `is_withdraw` is True, remove a BGP route from the Global table.
        """

        from os_ken.services.protocols.bgp.core import BgpCoreError
        from os_ken.services.protocols.bgp.api.prefix import (
            FLOWSPEC_FAMILY_IPV4,
            FLOWSPEC_FAMILY_IPV6,
            FLOWSPEC_FAMILY_L2VPN,
        )

        src_ver_num = 1
        peer = None

        # set mandatory path attributes
        origin = BGPPathAttributeOrigin(BGP_ATTR_ORIGIN_IGP)
        aspath = BGPPathAttributeAsPath([[]])

        pathattrs = OrderedDict()
        pathattrs[BGP_ATTR_TYPE_ORIGIN] = origin
        pathattrs[BGP_ATTR_TYPE_AS_PATH] = aspath

        if flowspec_family == FLOWSPEC_FAMILY_IPV4:
            _nlri = FlowSpecIPv4NLRI.from_user(**rules)
            p = IPv4FlowSpecPath

            try:
                communities = create_v4flowspec_actions(actions)
            except ValueError as e:
                raise BgpCoreError(desc=str(e))

            if communities:
                pathattrs[BGP_ATTR_TYPE_EXTENDED_COMMUNITIES] = (
                    BGPPathAttributeExtendedCommunities(
                        communities=communities))
        elif flowspec_family == FLOWSPEC_FAMILY_IPV6:
            _nlri = FlowSpecIPv6NLRI.from_user(**rules)
            p = IPv6FlowSpecPath

            try:
                communities = create_v6flowspec_actions(actions)
            except ValueError as e:
                raise BgpCoreError(desc=str(e))

            if communities:
                pathattrs[BGP_ATTR_TYPE_EXTENDED_COMMUNITIES] = (
                    BGPPathAttributeExtendedCommunities(
                        communities=communities))
        elif flowspec_family == FLOWSPEC_FAMILY_L2VPN:
            _nlri = FlowSpecL2VPNNLRI.from_user(**rules)
            p = L2vpnFlowSpecPath

            try:
                communities = create_l2vpnflowspec_actions(actions)
            except ValueError as e:
                raise BgpCoreError(desc=str(e))

            if communities:
                pathattrs[BGP_ATTR_TYPE_EXTENDED_COMMUNITIES] = (
                    BGPPathAttributeExtendedCommunities(
                        communities=communities))
        else:
            raise BgpCoreError(
                desc='Unsupported flowspec family %s' % flowspec_family)

        new_path = p(peer, _nlri, src_ver_num,
                     pattrs=pathattrs, is_withdraw=is_withdraw)

        # add to global table and propagates to neighbors
        self.learn_path(new_path)

    def clean_stale_routes(self, peer, route_family=None):
        """Removes old routes from `peer` from `route_family` table.

        Routes/paths version number is compared with `peer`s current version
        number.
        """

        if route_family is not None:
            if route_family not in SUPPORTED_GLOBAL_RF:
                raise ValueError('Given route family %s is not supported.' %
                                 route_family)

            tables = [self._global_tables.get(route_family)]
        else:
            tables = self._global_tables.values()
        for table in tables:
            table.cleanup_paths_for_peer(peer)



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2013 Isaku Yamahata <yamahata at private email ne jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Interface monitor.
Watching packet received on this interface and parse VRRP packet.

VRRPManager creates/deletes instances of interface monitor dynamically.
"""

from os_ken.base import app_manager
from os_ken.controller import handler
from os_ken.lib.packet import packet
from os_ken.lib.packet import vlan
from os_ken.lib.packet import vrrp
from os_ken.services.protocols.vrrp import event as vrrp_event


class VRRPInterfaceMonitor(app_manager.OSKenApp):
    # subclass of VRRPInterfaceBase -> subclass of VRRPInterfaceMonitor
    _CONSTRUCTORS = {}

    @staticmethod
    def register(interface_cls):
        def _register(cls):
            VRRPInterfaceMonitor._CONSTRUCTORS[interface_cls] = cls
            return cls
        return _register

    @staticmethod
    def factory(interface, config, router_name, statistics, *args, **kwargs):
        cls = VRRPInterfaceMonitor._CONSTRUCTORS[interface.__class__]
        app_mgr = app_manager.AppManager.get_instance()

        kwargs = kwargs.copy()
        kwargs['router_name'] = router_name
        kwargs['vrrp_config'] = config
        kwargs['vrrp_interface'] = interface
        kwargs['vrrp_statistics'] = statistics
        app = app_mgr.instantiate(cls, *args, **kwargs)
        return app

    @classmethod
    def instance_name(cls, interface, vrid):
        return '%s-%s-%d' % (cls.__name__, str(interface), vrid)

    def __init__(self, *args, **kwargs):
        super(VRRPInterfaceMonitor, self).__init__(*args, **kwargs)
        self.config = kwargs['vrrp_config']
        self.interface = kwargs['vrrp_interface']
        self.router_name = kwargs['router_name']
        self.statistics = kwargs['vrrp_statistics']
        self.name = self.instance_name(self.interface, self.config.vrid)

    def _parse_received_packet(self, packet_data):
        # OF doesn't support VRRP packet matching, so we have to parse
        # it ourselvs.
        packet_ = packet.Packet(packet_data)
        protocols = packet_.protocols

        # we expect either of
        #   [ether, vlan, ip, vrrp{, padding}]
        # or
        #   [ether, ip, vrrp{, padding}]

        if len(protocols) < 2:
            self.logger.debug('len(protocols) %d', len(protocols))
            return

        vlan_vid = self.interface.vlan_id
        may_vlan = protocols[1]
        if (vlan_vid is not None) != isinstance(may_vlan, vlan.vlan):
            self.logger.debug('vlan_vid: %s %s', vlan_vid, type(may_vlan))
            return
        if vlan_vid is not None and vlan_vid != may_vlan.vid:
            self.logger.debug('vlan_vid: %s vlan %s', vlan_vid, type(may_vlan))
            return

        # self.logger.debug('%s %s', packet_, packet_.protocols)
        may_ip, may_vrrp = vrrp.vrrp.get_payload(packet_)
        if not may_ip or not may_vrrp:
            # self.logger.debug('may_ip %s may_vrrp %s', may_ip, may_vrrp)
            return
        if not vrrp.vrrp.is_valid_ttl(may_ip):
            self.logger.debug('valid_ttl')
            return
        if may_vrrp.version != self.config.version:
            self.logger.debug('vrrp version %d %d',
                              may_vrrp.version, self.config.version)
            return
        if not may_vrrp.is_valid():
            self.logger.debug('valid vrrp')
            return
        offset = 0
        for proto in packet_.protocols:
            if proto == may_vrrp:
                break
            offset += len(proto)
        if not may_vrrp.checksum_ok(
                may_ip, packet_.data[offset:offset + len(may_vrrp)]):
            self.logger.debug('bad checksum')
            return
        if may_vrrp.vrid != self.config.vrid:
            self.logger.debug('vrid %d %d', may_vrrp.vrid, self.config.vrid)
            return
        if may_vrrp.is_ipv6 != self.config.is_ipv6:
            self.logger.debug('is_ipv6 %s %s',
                              may_vrrp.is_ipv6, self.config.is_ipv6)
            return

        # TODO: Optional check rfc5798 7.1
        # may_vrrp.ip_addresses equals to self.config.ip_addresses
        if may_vrrp.priority == 0:
            self.statistics.rx_vrrp_zero_prio_packets += 1

        vrrp_received = vrrp_event.EventVRRPReceived(self.interface, packet_)
        self.send_event(self.router_name, vrrp_received)
        return True

    def _send_vrrp_packet_received(self, packet_data):
        valid = self._parse_received_packet(packet_data)
        if valid is True:
            self.statistics.rx_vrrp_packets += 1
        else:
            self.statistics.rx_vrrp_invalid_packets += 1

    @handler.set_ev_handler(vrrp_event.EventVRRPTransmitRequest)
    def vrrp_transmit_request_handler(self, ev):
        raise NotImplementedError()

    def _initialize(self):
        raise NotImplementedError()

    def _shutdown(self):
        raise NotImplementedError()

    @handler.set_ev_handler(vrrp_event.EventVRRPStateChanged)
    def vrrp_state_changed_handler(self, ev):
        assert ev.interface == self.interface

        if ev.new_state == vrrp_event.VRRP_STATE_INITIALIZE:
            # add/del packet in rule
            if ev.old_state:
                self._shutdown()
            else:
                self._initialize()
        elif ev.new_state in [vrrp_event.VRRP_STATE_BACKUP,
                              vrrp_event.VRRP_STATE_MASTER]:

            if ev.old_state == vrrp_event.VRRP_STATE_INITIALIZE:
                if ev.new_state == vrrp_event.VRRP_STATE_MASTER:
                    self.statistics.idle_to_master_transitions += 1
                else:
                    self.statistics.idle_to_backup_transitions += 1
            elif ev.old_state == vrrp_event.VRRP_STATE_MASTER:
                self.statistics.master_to_backup_transitions += 1
            else:
                self.statistics.backup_to_master_transitions += 1
        else:
            raise RuntimeError('unknown vrrp state %s' % ev.new_state)



# Copyright (C) 2017 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import unittest

from nose.tools import eq_, raises

from os_ken.lib.packet.bgp import (
    BGPFlowSpecTrafficRateCommunity,
    BGPFlowSpecTrafficActionCommunity,
    BGPFlowSpecRedirectCommunity,
    BGPFlowSpecTrafficMarkingCommunity,
    BGPFlowSpecVlanActionCommunity,
    BGPFlowSpecTPIDActionCommunity,
)

from os_ken.services.protocols.bgp.core import BgpCoreError
from os_ken.services.protocols.bgp.utils.bgp import create_v4flowspec_actions
from os_ken.services.protocols.bgp.utils.bgp import create_v6flowspec_actions
from os_ken.services.protocols.bgp.utils.bgp import create_l2vpnflowspec_actions


LOG = logging.getLogger(__name__)


class Test_Utils_BGP(unittest.TestCase):
    """
    Test case for os_ken.services.protocols.bgp.utils.bgp
    """

    def _test_create_v4flowspec_actions(self, actions, expected_communities):
        communities = create_v4flowspec_actions(actions)
        expected_communities.sort(key=lambda x: x.subtype)
        communities.sort(key=lambda x: x.subtype)
        eq_(str(expected_communities), str(communities))

    def test_create_v4flowspec_actions_all_actions(self):
        actions = {
            'traffic_rate': {
                'as_number': 0,
                'rate_info': 100.0,
            },
            'traffic_action': {
                'action': 3,
            },
            'redirect': {
                'as_number': 10,
                'local_administrator': 10,
            },
            'traffic_marking': {
                'dscp': 24,
            }
        }
        expected_communities = [
            BGPFlowSpecTrafficRateCommunity(as_number=0, rate_info=100.0),
            BGPFlowSpecTrafficActionCommunity(action=3),
            BGPFlowSpecRedirectCommunity(as_number=10, local_administrator=10),
            BGPFlowSpecTrafficMarkingCommunity(dscp=24),
        ]
        self._test_create_v4flowspec_actions(actions, expected_communities)

    def test_create_v4flowspec_actions_without_actions(self):
        actions = None
        expected_communities = []
        self._test_create_v4flowspec_actions(actions, expected_communities)

    @raises(ValueError)
    def test_create_v4flowspec_actions_not_exist_actions(self):
        actions = {
            'traffic_test': {
                'test': 10,
            },
        }
        expected_communities = []
        self._test_create_v4flowspec_actions(actions, expected_communities)

    def _test_create_v6flowspec_actions(self, actions, expected_communities):
        communities = create_v6flowspec_actions(actions)
        expected_communities.sort(key=lambda x: x.subtype)
        communities.sort(key=lambda x: x.subtype)
        eq_(str(expected_communities), str(communities))

    def test_create_v6flowspec_actions_all_actions(self):
        actions = {
            'traffic_rate': {
                'as_number': 0,
                'rate_info': 100.0,
            },
            'traffic_action': {
                'action': 3,
            },
            'redirect': {
                'as_number': 10,
                'local_administrator': 10,
            },
            'traffic_marking': {
                'dscp': 24,
            }
        }
        expected_communities = [
            BGPFlowSpecTrafficRateCommunity(as_number=0, rate_info=100.0),
            BGPFlowSpecTrafficActionCommunity(action=3),
            BGPFlowSpecRedirectCommunity(as_number=10, local_administrator=10),
            BGPFlowSpecTrafficMarkingCommunity(dscp=24),
        ]
        self._test_create_v6flowspec_actions(actions, expected_communities)

    def test_create_v6flowspec_actions_without_actions(self):
        actions = None
        expected_communities = []
        self._test_create_v6flowspec_actions(actions, expected_communities)

    @raises(ValueError)
    def test_create_v6flowspec_actions_not_exist_actions(self):
        actions = {
            'traffic_test': {
                'test': 10,
            },
        }
        expected_communities = []
        self._test_create_v6flowspec_actions(actions, expected_communities)

    def _test_create_l2vpnflowspec_actions(self, actions, expected_communities):
        communities = create_l2vpnflowspec_actions(actions)
        expected_communities.sort(key=lambda x: x.subtype)
        communities.sort(key=lambda x: x.subtype)
        eq_(str(expected_communities), str(communities))

    def test_create_l2vpnflowspec_actions_all_actions(self):
        actions = {
            'traffic_rate': {
                'as_number': 0,
                'rate_info': 100.0,
            },
            'traffic_action': {
                'action': 3,
            },
            'redirect': {
                'as_number': 10,
                'local_administrator': 10,
            },
            'traffic_marking': {
                'dscp': 24,
            },
            'vlan_action': {
                'actions_1': (BGPFlowSpecVlanActionCommunity.POP |
                              BGPFlowSpecVlanActionCommunity.SWAP),
                'vlan_1': 3000,
                'cos_1': 3,
                'actions_2': BGPFlowSpecVlanActionCommunity.PUSH,
                'vlan_2': 4000,
                'cos_2': 2,
            },
            'tpid_action': {
                'actions': (BGPFlowSpecTPIDActionCommunity.TI |
                            BGPFlowSpecTPIDActionCommunity.TO),
                'tpid_1': 5,
                'tpid_2': 6,
            }
        }
        expected_communities = [
            BGPFlowSpecTrafficRateCommunity(as_number=0, rate_info=100.0),
            BGPFlowSpecTrafficActionCommunity(action=3),
            BGPFlowSpecRedirectCommunity(as_number=10, local_administrator=10),
            BGPFlowSpecTrafficMarkingCommunity(dscp=24),
            BGPFlowSpecVlanActionCommunity(
                actions_1=(BGPFlowSpecVlanActionCommunity.POP |
                           BGPFlowSpecVlanActionCommunity.SWAP),
                vlan_1=3000,
                cos_1=3,
                actions_2=BGPFlowSpecVlanActionCommunity.PUSH,
                vlan_2=4000,
                cos_2=2,
            ),
            BGPFlowSpecTPIDActionCommunity(
                actions=(BGPFlowSpecTPIDActionCommunity.TI |
                         BGPFlowSpecTPIDActionCommunity.TO),
                tpid_1=5,
                tpid_2=6,
            ),
        ]
        self._test_create_l2vpnflowspec_actions(actions, expected_communities)

    def test_create_l2vpnflowspec_actions_without_actions(self):
        actions = None
        expected_communities = []
        self._test_create_l2vpnflowspec_actions(actions, expected_communities)

    @raises(ValueError)
    def test_create_l2vpnflowspec_actions_not_exist_actions(self):
        actions = {
            'traffic_test': {
                'test': 10,
            },
        }
        expected_communities = []
        self._test_create_l2vpnflowspec_actions(actions, expected_communities)



# Copyright (C) 2016 Bouygues Telecom.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
DHCPv6 packet parser/serializer

[RFC 3315] DHCPv6 packet format:

The following diagram illustrates the format of DHCP messages sent
between clients and servers::

     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |    msg_type   |               transaction_id                  |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                                                               |
    .                            options                            .
    .                           (variable)                          .
    |                                                               |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

There are two relay agent messages, which share the following format::

     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |    msg_type   |   hop_count   |                               |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+                               |
    |                                                               |
    |                         link_address                          |
    |                                                               |
    |                               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
    |                               |                               |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+                               |
    |                                                               |
    |                         peer_address                          |
    |                                                               |
    |                               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
    |                               |                               |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+                               |
    .                                                               .
    .            options (variable number and length)   ....        .
    |                                                               |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
"""
import random
import struct

from . import packet_base
from os_ken.lib import addrconv
from os_ken.lib import stringify

# DHCPv6 message types
DHCPV6_SOLICIT = 1
DHCPV6_ADVERTISE = 2
DHCPV6_REQUEST = 3
DHCPV6_CONFIRM = 4
DHCPV6_RENEW = 5
DHCPV6_REBIND = 6
DHCPV6_REPLY = 7
DHCPV6_RELEASE = 8
DHCPV6_DECLINE = 9
DHCPV6_RECONFIGURE = 10
DHCPV6_INFORMATION_REQUEST = 11
DHCPV6_RELAY_FORW = 12
DHCPV6_RELAY_REPL = 13

# DHCPv6 option-codes
DHCPV6_OPTION_CLIENTID = 1
DHCPV6_OPTION_SERVERID = 2
DHCPV6_OPTION_IA_NA = 3
DHCPV6_OPTION_IA_TA = 4
DHCPV6_OPTION_IAADDR = 5
DHCPV6_OPTION_ORO = 6
DHCPV6_OPTION_PREFERENCE = 7
DHCPV6_OPTION_ELAPSED_TIME = 8
DHCPV6_OPTION_RELAY_MSG = 9
DHCPV6_OPTION_AUTH = 11
DHCPV6_OPTION_UNICAST = 12
DHCPV6_OPTION_STATUS_CODE = 13
DHCPV6_OPTION_RAPID_COMMIT = 14
DHCPV6_OPTION_USER_CLASS = 15
DHCPV6_OPTION_VENDOR_CLASS = 16
DHCPV6_OPTION_VENDOR_OPTS = 17
DHCPV6_OPTION_INTERFACE_ID = 18
DHCPV6_OPTION_RECONF_MSG = 19
DHCPV6_OPTION_RECONF_ACCEPT = 20


class dhcp6(packet_base.PacketBase):
    """DHCPv6 (RFC 3315) header encoder/decoder class.

    The serialized packet would looks like the ones described
    in the following sections.

    * RFC 3315 DHCP packet format

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.


    ============== ====================
    Attribute      Description
    ============== ====================
    msg_type       Identifies the DHCP message type
    transaction_id For unrelayed messages only: the transaction ID for\
                   this message exchange.
    hop_count      For relayed messages only: number of relay agents that\
                   have relayed this message.
    link_address   For relayed messages only: a global or site-local address\
                   that will be used by the server to identify the link on\
                   which the client is located.
    peer_address   For relayed messages only: the address of the client or\
                   relay agent from which the message to be relayed was\
                   received.
    options        Options carried in this message
    ============== ====================
    """
    _MIN_LEN = 8
    _DHCPV6_UNPACK_STR = '!I'
    _DHCPV6_RELAY_UNPACK_STR = '!H16s16s'
    _DHCPV6_UNPACK_STR_LEN = struct.calcsize(_DHCPV6_UNPACK_STR)
    _DHCPV6_RELAY_UNPACK_STR_LEN = struct.calcsize(_DHCPV6_RELAY_UNPACK_STR)
    _DHCPV6_PACK_STR = '!I'
    _DHCPV6_RELAY_PACK_STR = '!H16s16s'

    def __init__(self, msg_type, options, transaction_id=None, hop_count=0,
                 link_address='::', peer_address='::'):
        super(dhcp6, self).__init__()
        self.msg_type = msg_type
        self.options = options
        if transaction_id is None:
            self.transaction_id = random.randint(0, 0xffffff)
        else:
            self.transaction_id = transaction_id
        self.hop_count = hop_count
        self.link_address = link_address
        self.peer_address = peer_address

    @classmethod
    def parser(cls, buf):
        (msg_type, ) = struct.unpack_from('!B', buf)

        buf = b'\x00' + buf[1:]  # unpack xid as a 4-byte integer
        if msg_type == DHCPV6_RELAY_FORW or msg_type == DHCPV6_RELAY_REPL:
            (hop_count, link_address, peer_address) \
                = struct.unpack_from(cls._DHCPV6_RELAY_UNPACK_STR, buf)
            length = struct.calcsize(cls._DHCPV6_RELAY_UNPACK_STR)
        else:
            (transaction_id, ) \
                = struct.unpack_from(cls._DHCPV6_UNPACK_STR, buf)
            length = struct.calcsize(cls._DHCPV6_UNPACK_STR)

        if len(buf) > length:
            parse_opt = options.parser(buf[length:])
            length += parse_opt.options_len
            if msg_type == DHCPV6_RELAY_FORW or msg_type == DHCPV6_RELAY_REPL:
                return (cls(msg_type, parse_opt, 0, hop_count,
                            addrconv.ipv6.bin_to_text(link_address),
                            addrconv.ipv6.bin_to_text(peer_address)),
                        None, buf[length:])
            else:
                return (cls(msg_type, parse_opt, transaction_id),
                        None, buf[length:])
        else:
            return None, None, buf

    def serialize(self, payload=None, prev=None):
        seri_opt = self.options.serialize()
        if (self.msg_type == DHCPV6_RELAY_FORW or
                self.msg_type == DHCPV6_RELAY_REPL):
            pack_str = '%s%ds' % (self._DHCPV6_RELAY_PACK_STR,
                                  self.options.options_len)
            buf = struct.pack(pack_str, self.hop_count,
                              addrconv.ipv6.text_to_bin(self.link_address),
                              addrconv.ipv6.text_to_bin(self.peer_address),
                              seri_opt)
        else:
            pack_str = '%s%ds' % (self._DHCPV6_PACK_STR,
                                  self.options.options_len)
            buf = struct.pack(pack_str, self.transaction_id, seri_opt)
        return struct.pack('!B', self.msg_type) + buf[1:]


class options(stringify.StringifyMixin):
    """DHCP (RFC 3315) options encoder/decoder class.

    This is used with os_ken.lib.packet.dhcp6.dhcp6.
    """

    def __init__(self, option_list=None, options_len=0):
        super(options, self).__init__()
        if option_list is None:
            self.option_list = []
        else:
            self.option_list = option_list
        self.options_len = options_len

    @classmethod
    def parser(cls, buf):
        opt_parse_list = []
        offset = 0
        while len(buf) > offset:
            opt_buf = buf[offset:]
            opt = option.parser(opt_buf)
            opt_parse_list.append(opt)
            offset += opt.length + 4
        return cls(opt_parse_list, len(buf))

    def serialize(self):
        seri_opt = bytes()
        for opt in self.option_list:
            seri_opt += opt.serialize()
        if self.options_len == 0:
            self.options_len = len(seri_opt)
        return seri_opt


class option(stringify.StringifyMixin):
    """DHCP (RFC 3315) options encoder/decoder class.

    This is used with os_ken.lib.packet.dhcp6.dhcp6.options.

    An instance has the following attributes at least.
    Most of them are same to the on-wire counterparts but in host byte order.
    __init__ takes the corresponding args in this order.

    The format of DHCP options is::

         0                   1                   2                   3
         0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |          option-code          |           option-len          |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                          option-data                          |
        |                      (option-len octets)                      |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    ============== ====================
    Attribute      Description
    ============== ====================
    option-code    An unsigned integer identifying the specific option\
                   type carried in this option.
    option-len     An unsigned integer giving the length of the\
                   option-data field in this option in octets.
    option-data    The data for the option; the format of this data\
                   depends on the definition of the option.
    ============== ====================
    """
    _UNPACK_STR = '!H'
    _UNPACK_STR_LEN = struct.calcsize(_UNPACK_STR)
    _PACK_STR = '!HH%ds'

    def __init__(self, code, data, length=0):
        super(option, self).__init__()
        self.code = code
        self.data = data
        self.length = length

    @classmethod
    def parser(cls, buf):
        code = struct.unpack_from(cls._UNPACK_STR, buf)[0]
        buf = buf[cls._UNPACK_STR_LEN:]
        length = struct.unpack_from(cls._UNPACK_STR, buf)[0]
        buf = buf[cls._UNPACK_STR_LEN:]
        value_unpack_str = '%ds' % length
        data = struct.unpack_from(value_unpack_str, buf)[0]
        return cls(code, data, length)

    def serialize(self):
        if self.length == 0:
            self.length = len(self.data)
        options_pack_str = self._PACK_STR % self.length
        return struct.pack(options_pack_str, self.code, self.length, self.data)
